import collections.abc
import logging
import xml.etree.ElementTree as ET
from io import BytesIO
from pathlib import Path
from time import perf_counter as now
from typing import Dict, Iterator, List, Tuple, Union

import pandas as pd
from numpy import nan
from pandas import DataFrame

from lac_validator.config import column_names
from lac_validator.datastore import la_df, merge_postcodes
from lac_validator.types import UploadedFile, UploadError

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # uncomment if you want to see debug messages


class Timer:
    def __init__(self):
        self._prev = None
        self._zero = None
        self.t0

    @property
    def t0(self):
        self._zero = now()
        self._prev = self._zero
        return f"Starting timer"

    @property
    def t(self):
        t = now()
        elapsed = t - self._prev
        self._prev = t
        return f"{elapsed:05.2f}s (total {t - self._zero:05.02f}s)"


sc = Timer()


class _BufferedUploadedFile(collections.abc.Mapping):
    def __init__(self, file, name, description):
        self.name = name
        self.description = description
        self.file = Path(file)
        if not self.file.is_file():
            raise FileNotFoundError(f"{self.file} not found.")

    def __getitem__(self, k):
        if k == "name":
            return self.name
        elif k == "description":
            return self.description
        elif k == "file_content":
            with open(self.file, "rb") as file:
                return file.read()
        else:
            raise AttributeError(f"{k} not found")

    def __len__(self) -> int:
        return 3

    def __iter__(self) -> Iterator:
        pass


def read_from_text(
    raw_files: List[UploadedFile],
) -> Tuple[Dict[str, DataFrame], Dict[str, Union[str, DataFrame]]]:
    """
    Reads from a raw list of files passed from javascript. These files are of
    the form e.g.
    [
        {name: 'filename.csv', file_content: <file contents>, description: <upload metadata>}
    ]

    This function will try to catch most basic upload errors, and dispatch other errors
    to either the csv or xml reader based on the file extension.
    """
    logger.info(f"Reading from text. {sc.t0}")
    metadata_extras = {}

    CH_uploaded = [f for f in raw_files if f["description"] == "CH lookup"]
    SCP_uploaded = [f for f in raw_files if f["description"] == "SCP lookup"]
    num_of_CH_and_SCP = (len(CH_uploaded), len(SCP_uploaded))

    logger.info(f"#CH: {num_of_CH_and_SCP[0]}, #SCP:{num_of_CH_and_SCP[1]}")

    if max(num_of_CH_and_SCP) > 1:
        raise UploadError(
            f"{num_of_CH_and_SCP[0]} (Children's Homes List) and {num_of_CH_and_SCP[1]} (Social Care Providers List) "
            f"URN lookup tables were loaded - Please only load a single file in each box."
        )
    elif set(num_of_CH_and_SCP) == {0, 1}:
        if CH_uploaded:
            # Checks if a single CH list has wrongly been uploaded, or if a single combined SCPCH list has been uploaded
            combined_scpch = combined_ch_scp_check(CH_uploaded[0])
            if combined_scpch:
                logger.info(
                    f"Combined 'Childrens home' and 'Social Care Providers' lists detected. {sc.t}"
                )
                provider_info_df = scpch_provider_info_table(scpch=CH_uploaded[0])
                metadata_extras["provider_info"] = provider_info_df
            else:
                raise UploadError(
                    "Please load both the latest 'Children's Homes' and 'Social Care Providers' lists "
                    "from Ofsted into their respective boxes above."
                )
        else:
            raise UploadError(
                "Please load both the latest 'Children's Homes' and 'Social Care Providers' lists "
                "from Ofsted into their respective boxes above."
            )
    elif set(num_of_CH_and_SCP) == {1}:
        logger.info(
            f"Ofsted CH & SCP spreadsheets received - constructing provider info table. {sc.t}"
        )
        (CH_uploaded,) = CH_uploaded  # These should both be single-element lists
        (SCP_uploaded,) = SCP_uploaded
        provider_info_df = construct_provider_info_table(
            CH=CH_uploaded, SCP=SCP_uploaded
        )
        metadata_extras["provider_info"] = provider_info_df

        # logger.info(
        #     "created providers table. shape:{1} rows, {0}".format(
        #         *provider_info_df.shape
        #     )
        # )
        logger.debug(f'{", ".join(provider_info_df.columns)}')
        logger.debug(f"providers dtypes: {provider_info_df.dtypes}")

    else:
        logger.info(
            "Ofsted CH & SCP spreadsheets not loaded - checks involving URN lookup will be skipped."
        )

    raw_files = [
        f for f in raw_files if f["description"] not in ("CH lookup", "SCP lookup")
    ]

    extensions = list(set([f["name"].split(".")[-1].lower() for f in raw_files]))

    if len(raw_files) == 0:
        raise UploadError("No SSDA 903 files uploaded!")
    elif len(extensions) > 1:
        raise UploadError(
            f"Mix of CSV and XML files found ({extensions})! Please reupload."
        )
    else:
        if extensions == ["csv"]:
            metadata_extras["file_format"] = "csv"
            return read_csvs_from_text(raw_files), metadata_extras
        elif extensions == ["xml"]:
            metadata_extras["file_format"] = "xml"
            return read_xml_from_text(raw_files[0]["file_content"]), metadata_extras
        else:
            raise UploadError(f"Unknown file type {extensions[0]} found.")


def construct_provider_info_table(CH: UploadedFile, SCP: UploadedFile):
    """
    inputs:
    CH (childrens homes) and SCP (social care providers) are either strings containing a path to
    the excel files, or the files' contents as passed in by the frontend

    returns:
    provider_info_df is a dataframe
    (*) it will be stored in dfs['metadata']['provider_info'] if it has been uploaded
    (*) if it's not there, then  return {} for rules requiring it
    (*) the columns will be ['URN', 'PLACE_CODES', 'PROVIDER_CODES', 'REG_END', 'POSTCODE', 'LA_CODE']:
    (*) all cols will be strings (object dtype) except REG_END which will be datetime
        - URN:
            merge on this.
        - PLACE_CODES and PROVIDER_CODES:
            strings of codes separated by commas: ['PR4', 'PR1,PR2', etc..]
        - REG_END:
            date provider should have closed (usually blank) - note its type will already be datetime64
        - POSTCODE:
            postcode - upper-cased, spaces removed
        - LA_CODE_INFERRED and LA_NAME_INFERRED:
            utla code and utla name - obtained from la_df by matching on provider postcode
            (!) These may differ from the LA named in the files, either due to formatting, or if they serve
            different LA to where they are located. Rule descriptions should mention this caveat.
        - LA_NAME_FROM_FILE
            LA name from uploaded files.
    """
    if not isinstance(CH, str):
        if not isinstance(CH["file_content"], bytes):
            CH_bytes = CH["file_content"].tobytes()
        else:
            CH_bytes = CH["file_content"]
    else:
        CH_bytes = CH
    if not isinstance(SCP, str):
        if not isinstance(SCP["file_content"], bytes):
            SCP_bytes = SCP["file_content"].tobytes()
        else:
            SCP_bytes = SCP["file_content"]
    else:
        SCP_bytes = SCP
    provider_info_cols = [
        "URN",
        "LA_NAME_FROM_FILE",
        "PLACE_CODES",
        "PROVIDER_CODES",
        "REG_END",
        "POSTCODE",
    ]
    logger.info(f"URN lookup bytes recieved. Reading excel files... {sc.t}")

    # read childrens homes file
    CH_sheets = pd.ExcelFile(CH_bytes, engine="openpyxl").sheet_names
    CH_cols = [
        "URN",
        "Local Authority",
        "Provider Code",
        "Provider Placement Code",
        "Closed Date",
        "Setting Postcode",
    ]

    # check whether file includes consolidated provider information sheet
    if "Provider information" in CH_sheets:
        CH_providers = pd.read_excel(
            CH_bytes, sheet_name="Provider information", engine="openpyxl"
        )
        logger.debug(
            f"Reading CH provider info from excel done. cols:{CH_providers.columns} {sc.t}"
        )

        try:
            CH_address = CH_providers[["URN", "Setting Postcode"]]
        except KeyError:
            try:
                CH_address = CH_providers[["URN"]]
                CH_providers = CH_providers.rename(
                    columns={"Setting Address Postcode": "Setting Postcode"}
                )
            except KeyError:
                raise print(
                    "Failed to find postcode column in Childrens Homes list "
                    '"Provider information" sheet. Expected "Setting Postcode"'
                )
        try:
            CH_providers = CH_providers[CH_cols[:]]
        except KeyError:
            raise UploadError(
                f'Failed to find required columns in Childrens Homes list "Provider information"'
                f' sheet. Expected: {", ".join(CH_cols[:])}'
            )
        CH_df = CH_providers
        del CH_providers, CH_address

    # if not check whether file includes separate setting and address sheets
    elif "Settings and Inspection Info" in CH_sheets and "Address Details" in CH_sheets:
        # this sheet contains all columns except the postcode
        CH_setting = pd.read_excel(
            CH_bytes, sheet_name="Settings and Inspection Info", engine="openpyxl"
        )
        logger.debug(
            f"Reading CH setting info from excel done. cols:{CH_setting.columns} {sc.t}"
        )
        # from this sheet we need only the postcode
        CH_address = pd.read_excel(
            CH_bytes, sheet_name="Address Details", engine="openpyxl"
        )
        logger.debug(
            f"Reading CH address info from excel done. cols:{CH_address.columns} {sc.t}"
        )

        try:
            CH_address = CH_address[["URN", "Setting Postcode"]]
        except KeyError:
            try:
                CH_address = CH_address[["URN", "Setting Address Postcode"]].rename(
                    columns={"Setting Address Postcode": "Setting Postcode"}
                )
            except KeyError:
                raise UploadError(
                    "Failed to find postcode column in Childrens Homes list "
                    '"Address Details" sheet. Expected "Setting Postcode"'
                )
        try:
            CH_setting = CH_setting[CH_cols[:-1]]
        except KeyError:
            raise UploadError(
                f'Failed to find required columns in Childrens Homes list "Setting and Inspection Info"'
                f' sheet. Expected: {", ".join(CH_cols[: -1])}'
            )

        # combine the postcode with the rest of the columns
        CH_df = pd.merge(left=CH_setting, right=CH_address, on="URN", how="inner")
        del CH_setting, CH_address

    else:
        raise UploadError(
            "Failed to find expected sheet names in Childrens Homes List. "
            'Expected "Provider information" or "Setting and Inspection Info" and "Address Details"'
        )

    CH_df["source"] = "CH List"
    CH_df["Provider Placement Code"] = CH_df["Provider Placement Code"].str.replace(
        "/", ","
    )
    provider_info_df = CH_df.rename(columns=dict(zip(CH_cols, provider_info_cols)))
    del CH_df
    logger.info(f"CH dataframe complete. Creating SCP dataframe {sc.t}")

    # read social care providers file
    SCP_current = pd.read_excel(
        SCP_bytes, engine="openpyxl", sheet_name=0
    )  # current providers
    logger.debug(
        f"Reading SCP_current from excel done. cols: {SCP_current.columns}. {sc.t}"
    )
    SCP_closed = pd.read_excel(
        SCP_bytes, engine="openpyxl", sheet_name=1
    )  # closed providers
    logger.debug(
        f"Reading SCP_closed from excel done. cols: {SCP_closed.columns}. {sc.t}"
    )

    SCP_cols = [
        "URN",
        "Local authority",
        "Placement code",
        "Placement provider code",
        "Deregistration date",
        "Setting postcode",
    ]
    try:
        SCP_current = SCP_current[[_ for _ in SCP_cols if _ != "Deregistration date"]]
    except KeyError:
        missing_cols = set(_ for _ in SCP_cols if _ != "Deregistration date") - set(
            SCP_current.columns
        )
        raise UploadError(
            f'Failed to find required columns in Social Care Providers list "Current providers"'
            f' sheet. Didnt find: {", ".join(missing_cols)}'
        )
    try:
        SCP_closed = SCP_closed[SCP_cols]
    except KeyError:
        missing_cols = set(SCP_cols) - set(SCP_closed.columns)
        raise UploadError(
            f'Failed to find required columns in Social Care Providers list "Providers closed"'
            f' sheet. Didnt find: {", ".join(missing_cols)}'
        )

    # concatenate the two SCP sheets and rename the columns
    SCP_df = pd.concat((SCP_current, SCP_closed)).rename(
        columns=dict(zip(SCP_cols, provider_info_cols))
    )
    SCP_df = SCP_df[SCP_df["URN"].notnull()]
    SCP_df["source"] = "SCP List"

    # add SCP info to provider_info_df
    provider_info_df = pd.concat((provider_info_df, SCP_df), ignore_index=True)
    del SCP_current, SCP_closed, SCP_df

    # standardise postcodes
    provider_info_df["POSTCODE"] = (
        provider_info_df["POSTCODE"].str.replace(" ", "").str.upper()
    )

    # infer LA based on provider's postcode. this does not necessarily match what's in the file!
    provider_info_df[["LA_CODE_INFERRED", "LA_NAME_INFERRED"]] = (
        merge_postcodes(provider_info_df, "POSTCODE")
        .merge(la_df, how="left", left_on="laua", right_on="LTLA21CD")
        .loc[:, ["UTLA21CD", "UTLA21NM"]]
    )

    logger.info(f"Provider info dataframe successfully created {sc.t}")
    return provider_info_df


def scpch_provider_info_table(scpch: UploadedFile):
    """
    inputs:
    Combined CH (childrens homes) and SCP (social care providers) lists as the files' contents as passed in by the frontend

    returns:
    provider_info_df is a dataframe
    (*) it will be stored in dfs['metadata']['provider_info'] if it has been uploaded
    (*) if it's not there, then  return {} for rules requiring it
    (*) the columns will be ['URN', 'PLACE_CODES', 'PROVIDER_CODES', 'REG_END', 'POSTCODE', 'LA_CODE']:
    (*) all cols will be strings (object dtype) except REG_END which will be datetime
        - URN:
            URN of providers/homes, used for merging if there is more htan one list.
        - PLACE_CODES and PROVIDER_CODES:
            strings of codes separated by commas: ['PR4', 'PR1,PR2', etc..]
        - REG_END:
            date provider should have closed (usually blank) - note its type will already be datetime64
        - POSTCODE:
            postcode - upper-cased, spaces removed
        - LA_CODE_INFERRED and LA_NAME_INFERRED:
            utla code and utla name - obtained from la_df by matching on provider postcode
            (!) These may differ from the LA named in the files, either due to formatting, or if they serve
            different LA to where they are located. Rule descriptions should mention this caveat.
        - LA_NAME_FROM_FILE
            LA name from uploaded files.
    """
    if not isinstance(scpch, str):
        if not isinstance(scpch["file_content"], bytes):
            scpch_bytes = scpch["file_content"].tobytes()
        else:
            scpch_bytes = scpch["file_content"]
    else:
        scpch_bytes = scpch

    provider_info_cols = [
        "URN",
        "LA_NAME_FROM_FILE",
        "PLACE_CODES",
        "PROVIDER_CODES",
        "REG_END",
        "POSTCODE",
    ]

    scpch_cols = [
        "urn",
        "local authority",
        "placement code",
        "placement provider code",
        "deregistration date",
        "setting address postcode",
    ]
    logger.info(f"URN lookup bytes recieved. Reading excel files... {sc.t}")

    scpch_providers = pd.read_excel(scpch_bytes, sheet_name=None, engine="openpyxl")
    # Checks to see if it's the one sheet or two sheet version
    if len(scpch_providers) == 1:
        # next iter has a lower memory overhead than extracting the first element via a list
        # when we don't want to hard code the sheet names
        scpch_providers = next(iter(scpch_providers.values()))
        scpch_providers.columns = scpch_providers.columns.str.lower()
    if len(scpch_providers) == 2:
        scpch_iter = iter(scpch_providers.values())
        scpch_1, scpch_2 = next(scpch_iter), next(scpch_iter)

        scpch_2.columns = scpch_2.columns.str.lower()
        scpch_1.columns = scpch_1.columns.str.lower()

        # Supported accomodation has the placement code K3 (I believe)
        scpch_2["placement code"] = "K3"

        scpch_2["deregistration date"] = scpch_2["closed date"]

        if "placement provider code" not in scpch_2.columns:
            scpch_2["placement provider code"] = "PR4"

        scpch_2 = scpch_2[
            [
                "urn",
                "provider local authority",
                "placement code",
                "placement provider code",
                "deregistration date",
                "setting address postcode",
            ]
        ]
        scpch_providers = pd.concat([scpch_1, scpch_2])

    logger.debug(
        f"Reading SCP/CH provider info from excel done. cols:{scpch_providers.columns} {sc.t}"
    )

    try:
        scpch_providers = scpch_providers[scpch_cols[:]]
    except KeyError:
        raise UploadError(
            f'Failed to find required columns in Childrens Homes list "Provider information"'
            f' sheet. Expected: {", ".join(scpch_cols[:])}'
        )
    scpch_df = scpch_providers
    del scpch_providers

    scpch_df["source"] = "Combined SCPCH List"
    scpch_df["placement provider code"] = scpch_df[
        "placement provider code"
    ].str.replace("/", ",")
    provider_info_df = scpch_df.rename(
        columns=dict(zip(scpch_cols, provider_info_cols))
    )
    del scpch_df
    logger.info(f"CH dataframe complete. Creating SCP dataframe {sc.t}")

    # standardise postcodes
    provider_info_df["POSTCODE"] = (
        provider_info_df["POSTCODE"].str.replace(" ", "").str.upper()
    )

    # infer LA based on provider's postcode. this does not necessarily match what's in the file!
    provider_info_df[["LA_CODE_INFERRED", "LA_NAME_INFERRED"]] = (
        merge_postcodes(provider_info_df, "POSTCODE")
        .merge(la_df, how="left", left_on="laua", right_on="LTLA21CD")
        .loc[:, ["UTLA21CD", "UTLA21NM"]]
    )

    logger.info(f"Provider info dataframe successfully created {sc.t}")
    return provider_info_df


def read_files(files: Union[str, Path]) -> List[UploadedFile]:
    uploaded_files: List[_BufferedUploadedFile] = []
    for filename in files:
        uploaded_files.append(
            _BufferedUploadedFile(file=filename, name=filename, description="This year")
        )
    return uploaded_files


def capitalise_object_dtype_cols(df) -> pd.DataFrame:
    """This function takes in a pandas dataframe and capitalizes all the strings found in it."""
    for col in df.select_dtypes(include="object"):
        df[col] = df[col].str.upper()
    return df


def all_cols_to_object_dtype(df) -> pd.DataFrame:
    """This function converts all columns to object dtype."""
    for col in df.columns:
        if df.dtypes[col] != object:
            df[col] = df[col].values.astype(object)
    return df


def read_csvs_from_text(raw_files: List[UploadedFile]) -> Dict[str, DataFrame]:
    def _get_file_type(df) -> str:
        for table_name, expected_columns in column_names.items():
            if set(df.columns) == set(expected_columns):
                logger.info(f"Loaded {table_name} from CSV. ({len(df)} rows)")
                return table_name
        else:
            raise UploadError(
                f"Failed to match provided data ({list(df.columns)}) to known column names!"
            )

    files = {}
    for file_data in raw_files:
        csv_file = BytesIO(file_data["file_content"])
        # pd.read_csv on utf-16 files will raise a UnicodeDecodeError. This block prints a descriptive error message if that happens.
        try:
            max_cols = max([len(cols) for cols in column_names.values()])
            df = pd.read_csv(
                csv_file,
                converters={
                    i: lambda s: str(s) if s != "" else nan for i in range(max_cols)
                },
            )

        except UnicodeDecodeError:
            # raw_files is a list of files of type UploadedFile(TypedDict) whose instance is a dictionary containing the fields name, file_content, Description.
            # TODO: attempt to identify files that couldnt be decoded at this point; continue; then raise the exception outside the for loop, naming the uploaded filenames
            raise UploadError(
                f"Failed to decode one or more files. Try opening the text "
                f"file(s) in Notepad, then 'Saving As...' with the UTF-8 encoding"
            )
        # arrange column data types
        logger.debug("+" * 50)
        logger.debug("DF DATATYPES BEFORE CONVERSION", df.dtypes)
        df = all_cols_to_object_dtype(df)
        logger.debug("AFTER CONVERSION BEFORE CAPITALISING", df.dtypes)
        # capitalize all string input
        df = capitalise_object_dtype_cols(df)
        logger.debug("DF DATATYPES AFTER CAPITALISING", df.dtypes)

        file_name = _get_file_type(df)

        if "This year" in file_data["description"]:
            name = file_name
        elif "Prev year" in file_data["description"]:
            name = file_name + "_last"
        else:
            raise UploadError(
                f'Unrecognized file description {file_data["description"]}'
            )

        files[name] = df
        logger.debug("DF NAME: ", name)
        logger.debug("+" * 50)

    # Adding UASC column to Header table
    for header_name, uasc_name in (("Header", "UASC"), ("Header_last", "UASC_last")):
        if header_name in files and uasc_name in files:
            header = files[header_name]
            uasc = files[uasc_name]
            merge_indicator = header.merge(
                uasc, how="left", on="CHILD", indicator=True
            )["_merge"]

            header.loc[merge_indicator == "both", "UASC"] = "1"
            header.loc[merge_indicator == "left_only", "UASC"] = "0"

    return files


def read_xml_from_text(xml_string) -> Dict[str, DataFrame]:
    header_df = []
    episodes_df = []
    uasc_df = []
    oc2_df = []
    oc3_df = []
    ad1_df = []
    reviews_df = []
    sbpfa_df = []
    prev_perm_df = []
    missing_df = []

    def read_data(table):
        # The CHILDID tag needs to be renamed to CHILD to match the CSV
        # The PL tag needs to be renamed to PLACE to match the CSV
        conversions = {
            "CHILDID": "CHILD",
            "PL": "PLACE",
        }
        return {
            conversions.get(node.tag, node.tag): node.text
            for node in table.iter()
            if len(node) == 0
        }

    def get_fields_for_table(all_data, table_name):
        def read_value(k):
            val = all_data.get(k, None)
            if val is None:
                return nan
            if not isinstance(val, str):
                logger.warning(
                    f"Got a non-string thing reading {table_name}:{k} from xml -- got {val}, of type {type(val)}"
                )
            return val

        # Add UASC column to Header and Header_last tables
        cols = column_names[table_name]
        if table_name in ["Header", "Header_last"]:
            cols = cols + ["UASC"]

        return pd.Series({k: read_value(k) for k in cols}, dtype=object)

    for child in ET.fromstring(xml_string):
        all_data = read_data(child)
        header_df.append(get_fields_for_table(all_data, "Header"))
        if all_data.get("UASC", None) is not None:
            uasc_df.append(get_fields_for_table(all_data, "UASC"))
        if all_data.get("IN_TOUCH", None) is not None:
            oc3_df.append(get_fields_for_table(all_data, "OC3"))
        if all_data.get("DATE_INT", None) is not None:
            ad1_df.append(get_fields_for_table(all_data, "AD1"))
        for table in child:
            if table.tag == "EPISODE":
                data = read_data(table)
                episodes_df.append(
                    get_fields_for_table({**all_data, **data}, "Episodes")
                )
            elif table.tag == "HEADER":
                for child_table in table:
                    if child_table.tag == "AREVIEW":
                        data = read_data(child_table)
                        reviews_df.append(
                            get_fields_for_table({**all_data, **data}, "Reviews")
                        )
                    elif child_table.tag == "AMISSING":
                        data = read_data(child_table)
                        missing_df.append(
                            get_fields_for_table({**all_data, **data}, "Missing")
                        )
                    elif child_table.tag == "OC2":
                        data = read_data(child_table)
                        oc2_df.append(get_fields_for_table({**all_data, **data}, "OC2"))
                    elif child_table.tag == "PERMANENCE":
                        data = read_data(child_table)
                        prev_perm_df.append(
                            get_fields_for_table({**all_data, **data}, "PrevPerm")
                        )
                    elif child_table.tag == "AD_PLACED":
                        data = read_data(child_table)
                        sbpfa_df.append(
                            get_fields_for_table({**all_data, **data}, "PlacedAdoption")
                        )

    data = {
        "Header": pd.DataFrame(header_df),
        "Episodes": pd.DataFrame(episodes_df),
        "UASC": pd.DataFrame(uasc_df),
        "Reviews": pd.DataFrame(reviews_df),
        "OC2": pd.DataFrame(oc2_df),
        "OC3": pd.DataFrame(oc3_df),
        "AD1": pd.DataFrame(ad1_df),
        "PlacedAdoption": pd.DataFrame(sbpfa_df),
        "PrevPerm": pd.DataFrame(prev_perm_df),
        "Missing": pd.DataFrame(missing_df),
    }

    # capitalize string columns
    for df in data.values():
        df = all_cols_to_object_dtype(df)
        df = capitalise_object_dtype_cols(df)

    names_and_lengths = ", ".join(f"{t}: {len(data[t])} rows" for t in data)
    logger.info(f"Tables created from XML -- {names_and_lengths}")
    return data


def combined_ch_scp_check(excel_to_check):
    """
    Checks whether the file uploaded to the front end in the Children's home box is a CH list or a combined SPC/CH list.
    Only runs in instances where the number fo files uploaded in the SCP/CH boxes is one.
    """
    if isinstance(excel_to_check, bytes):
        CH_bytes = excel_to_check
    elif isinstance(excel_to_check, dict):
        if isinstance(excel_to_check["file_content"], bytes):
            CH_bytes = excel_to_check["file_content"]
        elif not isinstance(excel_to_check["file_content"], bytes):
            CH_bytes = excel_to_check["file_content"].tobytes()
    else:
        logger.info(
            f"Something is wrong with your 'Children's Home' or 'Social Care Providers' lists, check they're uploaded in the right boxes \
                    \n and that the column names are correct. {sc.t}"
        )
        raise UploadError(
            "Only one 'Children's Home' or 'Social Care Providers' list detected in upload,"
            "but it doesn't appear to be a combined list."
            "Please upload lists from Ofsted into their respective boxes above."
        )
    input_file = pd.read_excel(CH_bytes, engine="openpyxl", sheet_name=None)

    if (set(input_file.keys()) == set(["Providers", "SA providers"])) | (
        set(input_file.keys()) == set(["Providers", "SA Providers"])
    ):
        logger.info(
            f"'Childrens home' and 'Social Care Providers' lists detected across two sheets of Excel workbook. {sc.t}"
        )
        return True
    elif len(input_file.keys()) == 1:
        df = next(iter(input_file.values()))
        df.columns = df.columns.str.lower()
        if "provider type" not in df.columns:
            logger.info(
                f"Something is wrong with your 'Children's Home' or 'Social Care Providers' lists, check they're uploaded in the right boxes \
                        \n and that the column names are correct. {sc.t}"
            )
            raise UploadError(
                "Only one 'Children's Home' or 'Social Care Providers' list detected in upload,"
                "but it doesn't appear to be a combined list."
                "Please upload lists from Ofsted into their respective boxes above."
            )
        if (len(df["provider type"]) > 1) & (
            "Children's Home" in df["provider type"].values
        ):
            logger.info(
                f"Combined 'Childrens home' and 'Social Care Providers' lists detected. {sc.t}"
            )
            return True
        if (len(df["provider type"]) == 1) & (
            "Children's Home" in df["provider type"].values
        ):
            logger.info(
                f"'Children's home' list as only CH data file in upload detected. {sc.t}"
            )
            raise UploadError(
                "Only one 'Children's Home' or 'Social Care Providers' list detected in upload,"
                "but it doesn't appear to be a combined list."
                "Please upload lists from Ofsted into their respective boxes above."
            )
        else:
            logger.info(
                f"Something is wrong with your 'Children's Home' or 'Social Care Providers' lists, check they're uploaded in the right boxes \
                        \n and that the column names are correct. {sc.t}"
            )
            raise UploadError(
                "Only one 'Children's Home' or 'Social Care Providers' list detected in upload,"
                "but it doesn't appear to be a combined list."
                "Please upload lists from Ofsted into their respective boxes above."
            )
    else:
        logger.info(
            f"Something is wrong with your 'Children's Home' or 'Social Care Providers' lists, check they're of the correct file type. {sc.t}"
        )
        raise UploadError(
            "Only one 'Children's Home' or 'Social Care Providers' list detected in upload,"
            "but the file cannot be read for some reason. Check it's the correct file type."
        )



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="628",
    message="Motherhood details are not required for care leavers who have not been looked after during the year.",
    affected_fields=["MOTHER"],
    tables=["Episodes", "Header", "OC3"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs or "OC3" not in dfs:
        return {}
    else:
        hea = dfs["Header"]
        epi = dfs["Episodes"]
        oc3 = dfs["OC3"]

        hea = hea.reset_index()
        oc3_no_nulls = oc3[oc3[["IN_TOUCH", "ACTIV", "ACCOM"]].notna().any(axis=1)]

        hea_merge_epi = hea.merge(epi, how="left", on="CHILD", indicator=True)
        hea_not_in_epi = hea_merge_epi[hea_merge_epi["_merge"] == "left_only"]

        cohort_to_check = hea_not_in_epi.merge(oc3_no_nulls, how="inner", on="CHILD")
        error_cohort = cohort_to_check[cohort_to_check["MOTHER"].notna()]

        error_list = list(set(error_cohort["index"].to_list()))
        error_list.sort()
        return {"Header": error_list}


def test_validate():
    import pandas as pd

    fake_data_hea = pd.DataFrame(
        {
            "CHILD": ["0", "1", "2", "3", "4", "5", "6"],
            "MOTHER": [1, pd.NA, 0, 1, 0, 1, 1],  # 1 will pass as null
        }
    )
    fake_data_epi = pd.DataFrame(
        {
            "CHILD": [
                "a",
                "1",
                "3",
                "3",
                "3",
                "4",
                "5",
            ],  # So 0, 2 and 6 are the ones not in episodes
        }
    )
    fake_data_oc3 = pd.DataFrame(
        [
            {"CHILD": "0", "IN_TOUCH": "Whatever", "ACTIV": pd.NA, "ACCOM": "Whatever"},
            {
                "CHILD": "2",
                "IN_TOUCH": pd.NA,
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },  # All null values so 2 will pass
            {"CHILD": "6", "IN_TOUCH": "Whatever", "ACTIV": pd.NA, "ACCOM": pd.NA},
            {"CHILD": "5", "IN_TOUCH": "Whatever", "ACTIV": pd.NA, "ACCOM": pd.NA},
        ]
    )
    fake_dfs = {
        "Header": fake_data_hea,
        "Episodes": fake_data_epi,
        "OC3": fake_data_oc3,
    }

    result = validate(fake_dfs)

    assert result == {"Header": [0, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="180",
    message="Data entry for the strengths and difficulties questionnaire (SDQ) score is invalid.",
    affected_fields=["SDQ_SCORE"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]

        oc2["SDQ_SCORE_num"] = pd.to_numeric(oc2["SDQ_SCORE"], errors="coerce")

        error_mask = oc2["SDQ_SCORE"].notna() & ~oc2["SDQ_SCORE_num"].isin(range(41))

        validation_error_locations = oc2.index[error_mask]

        return {"OC2": validation_error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SDQ_SCORE": ["10", 41, "58", 72, "0", 40, 39.5, 20.0, pd.NA, "XX"],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [1, 2, 3, 6, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="131",
    message="Data entry for being in touch after leaving care is invalid.",
    affected_fields=["IN_TOUCH"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}

    care_leavers = dfs["OC3"]
    code_list = ["YES", "NO", "DIED", "REFU", "NREQ", "RHOM"]
    mask = care_leavers["IN_TOUCH"].isin(code_list)

    validation_error_mask = ~mask
    validation_error_locations = care_leavers.index[validation_error_mask]

    return {"OC3": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "IN_TOUCH": ["yes", "YES", 1, "REFUSE", "REFU", "", pd.NA],
        }
    )

    fake_dfs = {"OC3": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC3": [0, 2, 3, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="331",
    message="Date of matching child and adopter(s) should be the same as, or prior to, the date of placement of adoption.",
    affected_fields=["DATE_MATCH", "DECOM", "REC"],  # AD1  # Episodes
    tables=["AD1", "Episodes"],
)
def validate(dfs):
    if "AD1" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        adt = dfs["AD1"]
        eps = dfs["Episodes"]

        # Save indexes of each table so we can retreive the original positions in each for our error rows
        adt["AD1_index"] = adt.index
        eps["Episodes_index"] = eps.index

        adt["DATE_MATCH"] = pd.to_datetime(
            adt["DATE_MATCH"], format="%d/%m/%Y", errors="coerce"
        )
        eps["DECOM"] = pd.to_datetime(eps["DECOM"], format="%d/%m/%Y", errors="coerce")

        # Only keep the episodes where <Adopted> = 'Y'
        adoption_eps = eps[eps["REC"].isin(["E11", "E12"])]

        # Merge AD1 and Episodes so we can compare DATE_MATCH and DECOM
        adoption_eps = adoption_eps.merge(adt, on="CHILD")

        # A child cannot be placed for adoption before the child has been matched with prospective adopter(s).
        error_mask = adoption_eps["DATE_MATCH"] > adoption_eps["DECOM"]

        # Get the rows of each table where the dates clash
        AD1_errs = list(adoption_eps.loc[error_mask, "AD1_index"].unique())
        Episodes_errs = list(adoption_eps.loc[error_mask, "Episodes_index"].unique())

        return {"AD1": AD1_errs, "Episodes": Episodes_errs}


def test_validate():
    import pandas as pd

    fake_adt = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "66"],
            "DATE_MATCH": [
                "01/01/2020",
                "02/06/2020",
                "11/11/2020",
                pd.NA,
                "01/02/2020",
                "04/06/2020",
            ],
        }
    )
    fake_eps = pd.DataFrame(
        {
            "CHILD": [
                "111",
                "222",
                "333",
                "444",
                "555",
                "555",
                "555",
                "66",
                "66",
                "66",
                "66",
            ],
            "DECOM": [
                "01/01/2020",
                "01/06/2020",
                "12/12/2020",
                "01/02/2020",
                "01/01/2020",
                "01/02/2020",
                "05/06/2020",
                "01/01/2020",
                "01/02/2020",
                "05/06/2020",
                "06/07/2020",
            ],
            "REC": [
                "E12",
                "E11",
                "E12",
                "E11",
                "xX",
                "E11",
                "E12",
                "Xx",
                "oO",
                "E11",
                "Oo",
            ],
        }
    )

    fake_dfs = {"AD1": fake_adt, "Episodes": fake_eps}

    result = validate(fake_dfs)

    assert result == {"AD1": [1], "Episodes": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="174",
    message="Mother's child date of birth is recorded but gender shows that the child is a male.",
    affected_fields=["SEX", "MC_DOB"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]

        child_is_male = header["SEX"].astype(str) == "1"
        mc_dob_recorded = header["MC_DOB"].notna()

        error_mask = child_is_male & mc_dob_recorded

        validation_error_locations = header.index[error_mask]

        return {"Header": validation_error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SEX": ["1", "1", "2", "2", 1, 1],
            "MC_DOB": [pd.NA, "19/02/2010", pd.NA, "19/02/2010", pd.NA, "19/02/2010"],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1007",
    message="Care leaver information is not required for 17- or 18-year olds who are still looked after [on "
    "their 17th or 18th birthday.]",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "OC3"],
)
def validate(dfs):
    if "Episodes" not in dfs or "OC3" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        oc3 = dfs["OC3"]
        collection_end = dfs["metadata"]["collection_end"]
        # convert dates to datetime format
        oc3["DOB"] = pd.to_datetime(oc3["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(episodes["DEC"], format="%d/%m/%Y")
        episodes["DECOM"] = pd.to_datetime(episodes["DECOM"], format="%d/%m/%Y")

        # prepare to merge
        episodes.reset_index(inplace=True)
        oc3.reset_index(inplace=True)
        merged = episodes.merge(oc3, on="CHILD", how="left", suffixes=["_eps", "_oc3"])
        this_year = collection_end.year

        # If <DOB> < 19 and >= to 17 years prior to <COLLECTION_END_DATE> and current episode <DEC> and or <REC>
        # not provided then <IN_TOUCH>, <ACTIV> and <ACCOM> should not be provided
        check_age = (
            merged["DOB"] + pd.offsets.DateOffset(years=17) <= collection_end
        ) & (merged["DOB"] + pd.offsets.DateOffset(years=19) > collection_end)
        # That is, check that 17<=age<19

        merged["bday"] = merged["DOB"].apply(lambda x: x.replace(year=this_year))

        merged.loc[merged["bday"] > collection_end, "bday"] -= pd.DateOffset(years=1)

        in_care_on_bday = (merged["DECOM"] <= merged["bday"]) & (
            merged["DEC"] > merged["bday"]
        )

        # if either DEC or REC are absent
        mask = (
            check_age
            & in_care_on_bday
            & merged[["IN_TOUCH", "ACTIV", "ACCOM"]].notna().any(axis=1)
        )
        # Then raise an error if either IN_TOUCH, ACTIV, or ACCOM have been provided too

        # error locations
        oc3_error_locs = merged.loc[mask, "index_oc3"]
        episode_error_locs = merged.loc[mask, "index_eps"]
        # one to many join implies use .unique on the 'one'
        return {"OC3": oc3_error_locs.unique().tolist()}


def test_validate():
    import pandas as pd

    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": ["F", "G", "H", "I"],
            "DOB": ["01/02/2000", "01/09/1999", "01/02/2000", "01/09/1999"],
            "IN_TOUCH": ["notNA", "notNA", "notNA", pd.NA],
            "ACTIV": [pd.NA, pd.NA, pd.NA, pd.NA],
            "ACCOM": [pd.NA, pd.NA, pd.NA, pd.NA],
        }
    )
    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["F", "G", "H", "I"],
            "DECOM": ["01/02/2015", "01/09/2015", "01/02/2015", "01/09/2015"],
            "DEC": ["01/03/2017", "01/09/2016", "01/01/2017", "01/09/2017"],
        }
    )

    metadata = {"collection_end": "31/03/2017"}

    fake_dfs = {
        "Episodes": fake_data_episodes,
        "OC3": fake_data_oc3,
        "metadata": metadata,
    }

    result = validate(fake_dfs)
    assert result == {
        "OC3": [
            0,
        ]
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="407",
    message="Reason episode ceased is Special Guardianship Order, but child has reached age 18.",
    affected_fields=["DEC", "DOB", "REC"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        header = dfs["Header"]
        code_list = ["E45", "E46", "E47", "E48"]

        # convert dates to datetime format
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        # prepare to merge
        episodes.reset_index(inplace=True)
        header.reset_index(inplace=True)
        merged = episodes.merge(
            header, on="CHILD", how="left", suffixes=["_eps", "_er"]
        )

        # If <REC> = ‘E45’ or ‘E46’ or ‘E47’ or ‘E48’ then <DOB> must be < 18 years prior to <DEC>
        mask = merged["REC"].isin(code_list) & (
            merged["DOB"] + pd.offsets.DateOffset(years=18) < merged["DEC"]
        )
        # That is, raise error if DEC > DOB + 10years

        # error locations
        header_error_locs = merged.loc[mask, "index_er"]
        episode_error_locs = merged.loc[mask, "index_eps"]
        # one to many join implies use .unique on the 'one'
        return {
            "Episodes": episode_error_locs.tolist(),
            "Header": header_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {"CHILD": 101, "DEC": "01/01/2009", "REC": "E45"},  # 0 pass
            {"CHILD": 102, "DEC": "01/01/2001", "REC": "A3"},  # 1
            {"CHILD": 103, "DEC": "20/12/2001", "REC": "E15"},  # 2
            {"CHILD": 104, "DEC": "03/01/2019", "REC": "E46"},  # 3 fail
            {"CHILD": 105, "DEC": "03/04/2008", "REC": "E48"},  # 4 pass
            {"CHILD": 106, "DEC": "01/01/2002", "REC": "X2"},  # 5
            {"CHILD": 107, "DEC": "10/01/2002", "REC": "E11"},  # 6
            {"CHILD": 108, "DEC": "11/02/2020", "REC": "E48"},  # 7 fail
            {"CHILD": 109, "DEC": "25/01/2002", "REC": "X1"},  # 8
            {"CHILD": 110, "DEC": "25/01/2002", "REC": "E47"},  # 9
            {"CHILD": 111, "DEC": pd.NA, "REC": "E45"},  # 10 ignored by dropna
            {"CHILD": 112, "DEC": "25/01/2002", "REC": "E46"},  # 11 ignored by dropna
        ]
    )
    fake_data_header = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DOB": "01/01/2000",
            },  # 0
            {
                "CHILD": 102,
                "DOB": "01/01/2001",
            },  # 1
            {
                "CHILD": 103,
                "DOB": "20/12/2001",
            },  # 2
            {
                "CHILD": 104,
                "DOB": "01/01/2000",
            },  # 3
            {
                "CHILD": 105,
                "DOB": "03/01/2000",
            },  # 4
            {
                "CHILD": 106,
                "DOB": "01/01/2002",
            },  # 5
            {
                "CHILD": 107,
                "DOB": "10/01/2002",
            },  # 6
            {
                "CHILD": 108,
                "DOB": "11/01/2002",
            },  # 7
            {
                "CHILD": 109,
                "DOB": "25/01/2002",
            },  # 8
            {
                "CHILD": 110,
                "DOB": "25/01/2002",
            },  # 9
            {
                "CHILD": 111,
                "DOB": "25/01/2002",
            },  # 10
            {
                "CHILD": 112,
                "DOB": pd.NA,
            },  # 11
        ]
    )
    fake_dfs = {"Episodes": fake_data_eps, "Header": fake_data_header}

    result = validate(fake_dfs)
    assert result == {"Episodes": [3, 7], "Header": [3, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1015",
    message="Placement provider is own provision but child not placed in own LA.",
    affected_fields=["PL_LA"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        local_authority = dfs["metadata"]["localAuthority"]

        placement_fostering_or_adoption = df["PLACE"].isin(
            [
                "A3",
                "A4",
                "A5",
                "A6",
                "U1",
                "U2",
                "U3",
                "U4",
                "U5",
                "U6",
            ]
        )
        own_provision = df["PLACE_PROVIDER"].eq("PR1")
        is_short_term = df["LS"].isin(["V3", "V4"])
        is_pl_la = df["PL_LA"].eq(local_authority)

        checked_episodes = (
            ~placement_fostering_or_adoption & ~is_short_term & own_provision
        )
        checked_episodes = checked_episodes & df["LS"].notna() & df["PLACE"].notna()
        mask = checked_episodes & ~is_pl_la

        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"PLACE": "E1", "LS": "C1", "PLACE_PROVIDER": "PR1", "PL_LA": "auth"},
            {"PLACE": "E1", "LS": "C1", "PLACE_PROVIDER": "PR1", "PL_LA": "other"},
            {"PLACE": "U2", "LS": "C1", "PLACE_PROVIDER": "PR1", "PL_LA": "other"},
            {"PLACE": "E1", "LS": "V3", "PLACE_PROVIDER": "PR1", "PL_LA": "other"},
            {"PLACE": "E1", "LS": "C1", "PLACE_PROVIDER": "PR2", "PL_LA": "other"},
            {"PLACE": pd.NA, "LS": "C1", "PLACE_PROVIDER": "PR1", "PL_LA": "other"},
            {"PLACE": "E1", "LS": pd.NA, "PLACE_PROVIDER": "PR1", "PL_LA": "other"},
            {"PLACE": "E1", "LS": "C1", "PLACE_PROVIDER": pd.NA, "PL_LA": "other"},
        ]
    )

    metadata = {
        "localAuthority": "auth",
    }

    fake_dfs = {"Episodes": fake_data, "metadata": metadata}

    assert validate(fake_dfs) == {"Episodes": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="303",
    message="If date Unaccompanied Asylum-Seeking Child (UASC) status ceased is not null, UASC status must be coded 1.",
    affected_fields=["DUC", "UASC"],
    tables=["UASC", "Header"],
)
def validate(dfs):
    if ("UASC" not in dfs) or ("Header" not in dfs):
        return {}
    elif "UASC" not in dfs["Header"].columns:
        return {}
    else:
        uasc = dfs["UASC"]
        header = dfs["Header"]

        # merge
        uasc.reset_index(inplace=True)
        header.reset_index(inplace=True)

        merged = header.merge(uasc, how="left", on="CHILD", suffixes=["_er", "_sc"])

        merged["UASC"] = pd.to_numeric(merged["UASC"], errors="coerce")

        # If <DUC> provided, then <UASC> must be '1'
        error_mask = merged["DUC"].notna() & (merged["UASC"] != 1)

        uasc_error_locs = merged.loc[error_mask, "index_sc"]
        header_error_locs = merged.loc[error_mask, "index_er"]

        return {
            "UASC": uasc_error_locs.tolist(),
            "Header": header_error_locs.tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_uasc = pd.DataFrame(
        {
            "CHILD": [
                0,
                1,
                2,
                3,
                4,
                5,
            ],
            "DUC": [
                pd.NA,
                "04/04/2021",
                "01/06/2020",
                pd.NA,
                "10/04/2020",
                "01/03/2021",
            ],
        }
    )
    fake_data_header = pd.DataFrame(
        {
            "CHILD": [
                0,
                1,
                2,
                3,
                4,
                5,
            ],
            "UASC": [0, 1, 0, "1", "0", 1],
        }
    )

    fake_dfs = {"UASC": fake_data_uasc, "Header": fake_data_header}

    result = validate(fake_dfs)

    assert result == {"UASC": [2, 4], "Header": [2, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW01STG1",
    message="Child looked after on 31 March, but no social worker episode information is reported",
    affected_fields=["DEC"],
    tables=["SWEpisodes", "Episodes"],
)
def validate(dfs):
    if ("SWEpisodes" not in dfs) | ("Episodes" not in dfs):
        return {}
    else:
        SWE = dfs["SWEpisodes"]
        episodes = dfs["Episodes"]

        # If <CLA_AT_31_MARCH> then at least one instance of each of the following items should be provided:
        # <SW_ID>, <SW_DECOM>, <SW_REASON>

        episodes["index"] = episodes.index

        # DfE states children are <CLA_AT_31_MARCH> when
        # <DEC> of the current episode is null and <LS> not = ‘V3’, ‘V4’
        looked_after_31_mar = episodes[
            (episodes["DEC"].isna()) & ~(episodes["LS"].isin(["V3", "V4"]))
        ]

        merged_df = looked_after_31_mar.merge(SWE, on="CHILD", how="left")

        condition = (
            merged_df["SW_ID"].isna()
            | merged_df["SW_DECOM"].isna()
            | merged_df["SW_REASON"].isna()
        )

        errors = merged_df[condition]
        error_rows = errors["index"].tolist()

        return {"SWEpisodes": error_rows}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        [
            {"CHILD": "child1", "DEC": pd.NA, "LS": "xx"},
            {"CHILD": "child2", "DEC": pd.NA, "LS": "xx"},
            {"CHILD": "child3", "DEC": pd.NA, "LS": "xx"},
            {"CHILD": "child4", "DEC": pd.NA, "LS": "xx"},
            {"CHILD": "child5", "DEC": "01/01/2000", "LS": "xx"},
            {"CHILD": "child6", "DEC": pd.NA, "LS": "V4"},
            {"CHILD": "child8", "DEC": pd.NA, "LS": "xx"},
        ]
    )

    fake_swe = pd.DataFrame(
        [
            {"CHILD": "child1", "SW_ID": pd.NA, "SW_DECOM": "xx", "SW_REASON": "xx"},
            {"CHILD": "child2", "SW_ID": "xx", "SW_DECOM": pd.NA, "SW_REASON": "xx"},
            {"CHILD": "child3", "SW_ID": "xx", "SW_DECOM": "xx", "SW_REASON": pd.NA},
            {"CHILD": "child4", "SW_ID": "xx", "SW_DECOM": "xx", "SW_REASON": "xx"},
            {"CHILD": "child5", "SW_ID": "xx", "SW_DECOM": "xx", "SW_REASON": pd.NA},
            {"CHILD": "child6", "SW_ID": "xx", "SW_DECOM": "xx", "SW_REASON": pd.NA},
            {"CHILD": "child7", "SW_ID": "xx", "SW_DECOM": "xx", "SW_REASON": pd.NA},
        ]
    )

    fake_dfs = {"SWEpisodes": fake_swe, "Episodes": fake_epi}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [0, 1, 2, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="576",
    message="There is an open missing/away from placement without authorisation period in "
    + "last year’s return and there is no corresponding period recorded at the start of "
    + "this year.",
    affected_fields=["CHILD"],
    tables=["MIssing", "Missing_last"],
)
def validate(dfs):
    if "Missing" not in dfs or "Missing_last" not in dfs:
        return {}
    else:
        mis = dfs["Missing"]
        mis_l = dfs["Missing_last"]
        mis["MIS_START"] = pd.to_datetime(
            mis["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )
        mis_l["MIS_START"] = pd.to_datetime(
            mis_l["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )

        mis.reset_index(inplace=True)
        mis["MIS_START"].fillna(
            pd.to_datetime("01/01/2099", format="%d/%m/%Y", errors="coerce"),
            inplace=True,
        )
        min_mis = mis.groupby(["CHILD"])["MIS_START"].idxmin()
        mis = mis.loc[min_mis, :]

        open_mis_l = mis_l.query("MIS_END.isnull()")

        err_coh = mis.merge(open_mis_l, how="left", on="CHILD", suffixes=["", "_LAST"])
        err_coh = err_coh.query(
            "(MIS_START != MIS_START_LAST) & MIS_START_LAST.notnull()"
        )

        err_list = err_coh["index"].unique().tolist()
        err_list.sort()
        return {"Missing": err_list}


def test_validate():
    import pandas as pd

    fake_mis_l = pd.DataFrame(
        [
            {"CHILD": "111", "MIS_START": "07/02/2020", "MIS_END": "07/02/2020"},  # 0
            {"CHILD": "222", "MIS_START": "07/02/2020", "MIS_END": pd.NA},  # 1
            {"CHILD": "333", "MIS_START": "03/02/2020", "MIS_END": "07/02/2020"},  # 2
            {"CHILD": "444", "MIS_START": "07/02/2020", "MIS_END": pd.NA},  # 3
            {"CHILD": "555", "MIS_START": "01/02/2020", "MIS_END": pd.NA},  # 4
            {"CHILD": "666", "MIS_START": "13/02/2020", "MIS_END": "07/02/2020"},  # 5
        ]
    )
    fake_mis = pd.DataFrame(
        [
            {"CHILD": "111", "MIS_START": "07/02/2020"},  # 0
            {"CHILD": "222", "MIS_START": "08/02/2020"},  # 1 Fails
            {"CHILD": "333", "MIS_START": "03/02/2020"},  # 2
            {"CHILD": "444", "MIS_START": pd.NA},  # 3 Fails
            {"CHILD": "555", "MIS_START": "01/02/2020"},  # 4
            {"CHILD": "666", "MIS_START": "13/02/2020"},  # 5
        ]
    )
    fake_dfs = {"Missing_last": fake_mis_l, "Missing": fake_mis}

    assert validate(fake_dfs) == {"Missing": [1, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="411",
    message="Placement location code disagrees with LA of placement.",
    affected_fields=["PL_LOCATION"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        local_authority = dfs["metadata"]["localAuthority"]

        mask = df["PL_LOCATION"].eq("IN") & df["PL_LA"].ne(local_authority)

        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PL_LA": ["auth", "somewhere else", "auth", "auth", "somewhere else"],
            "PL_LOCATION": ["IN", "OUT", pd.NA, "OUT", "IN"],
        }
    )

    fake_dfs = {"Episodes": fake_data, "metadata": {"localAuthority": "auth"}}

    # Note 2 and 3 pass as the rule is specific
    # about only checking that 'IN' is set correctly
    assert validate(fake_dfs) == {"Episodes": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="531",
    message="A placement provider code of PR5 cannot be associated with placements P1.",
    affected_fields=["PLACE", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        error_mask = (epi["PLACE"] == "P1") & (epi["PLACE_PROVIDER"] == "PR5")
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["P1", "A3", "K1", "P1", "P1", "P1"],
            "PLACE_PROVIDER": ["PR5", "PR3", "PR4", "PR4", "PR5", "PRO"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="632",
    message="Date of previous permanence order not a valid value. NOTE: This rule may result in false negatives where the period of care started before the current collection year",
    affected_fields=["DATE_PERM", "DECOM"],
    tables=["Episodes", "PrevPerm"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PrevPerm" not in dfs:
        return {}
    else:
        import lac_validator.rules.rule_utils

        # function to check that date is of the right format

        episodes = dfs["Episodes"]
        prevperm = dfs["PrevPerm"]

        # convert dates from strings to appropriate format.
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        prevperm["DATE_PERM_dt"] = prevperm["DATE_PERM"].apply(
            lac_validator.rules.rule_utils.valid_date
        )

        # if nans aren't dropped, idxmin() won't work. we can do this since dropping nan DECOMs doesn't affect the rule logic.
        decom_only_eps = episodes[["CHILD", "DECOM"]].dropna()

        # select first episodes
        first_eps_idxs = decom_only_eps.groupby("CHILD")["DECOM"].idxmin()
        first_eps = episodes[episodes.index.isin(first_eps_idxs)]

        # prepare to merge
        first_eps.reset_index(inplace=True)
        prevperm.reset_index(inplace=True)
        merged = first_eps.merge(
            prevperm, on="CHILD", how="left", suffixes=["_eps", "_prev"]
        )

        # If provided <DATE_PERM> should be prior to <DECOM> and in a valid format and contain a valid date Format should be DD/MM/YYYY or one or more elements of the date can be replaced by ZZ if part of the date element is not known.
        mask = (merged["DATE_PERM_dt"] >= merged["DECOM"]) | (
            merged["DATE_PERM"].notna() & merged["DATE_PERM_dt"].isna()
        )

        # error locations
        prev_error_locs = merged.loc[mask, "index_prev"]
        eps_error_locs = merged.loc[mask, "index_eps"]

        return {
            "Episodes": eps_error_locs.tolist(),
            "PrevPerm": prev_error_locs.unique().tolist(),
        }


def test_validate():
    fake_data_prevperm = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "103",
                "104",
                "105",
                "106",
                "107",
                "108",
                "109",
                "110",
            ],
            "DATE_PERM": [
                "xx/10/2011",
                "17/06/2001",
                "01/05/2000",
                pd.NA,
                "05/ZZ/2020",
                "ZZ/05/2021",
                "ZZ/ZZ/ZZZZ",
                "ZZ/ZZ/1993",
                "ZZ/ZZ/ZZ",
                "01/13/2000",
            ],
        }
    )

    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "101",
                "DECOM": "20/10/2011",
            },  # 0 - fail! DATE_PERM is not in appropriate format
            {
                "CHILD": "102",
                "DECOM": "19/11/2021",
            },  # 1 pass
            {
                "CHILD": "102",
                "DECOM": "17/06/2001",
            },  # 2 - fail! DATE_PERM == DECOM
            {
                "CHILD": "103",
                "DECOM": pd.NA,
            },  # 3 ignore DECOM is nan
            {
                "CHILD": "104",
                "DECOM": "11/09/2015",
            },  # 4 pass DATE_PERM is nan
            {
                "CHILD": "105",
                "DECOM": "01/03/2021",
            },  # 5 - Pass
            {
                "CHILD": "106",
                "DECOM": "01/07/2021",
            },  # 6 pass
            {
                "CHILD": "107",
                "DECOM": "01/07/2021",
            },  # 7 pass
            {
                "CHILD": "108",
                "DECOM": "01/07/2021",
            },  # 8 pass
            {
                "CHILD": "109",
                "DECOM": "01/07/2021",
            },  # 9 - fail! wrong zeds in DATE_PERM
            {
                "CHILD": "110",
                "DECOM": "01/07/2021",
            },  # 10 - fail! wrong month in DATE_PERM
        ]
    )

    fake_data_episodes["DECOM"] = pd.to_datetime(
        fake_data_episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
    )

    fake_dfs = {"PrevPerm": fake_data_prevperm, "Episodes": fake_data_episodes}

    result = validate(fake_dfs)
    # desired
    assert result == {"Episodes": [0, 2, 9, 10], "PrevPerm": [0, 1, 8, 9]}



import pandas as pd


def decom_before_dob(dfs, p_code, y_gap):
    epi = dfs["Episodes"]
    hea = dfs["Header"]

    hea["DOB"] = pd.to_datetime(hea["DOB"], format="%d/%m/%Y", errors="coerce")
    epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
    epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")

    epi.reset_index(inplace=True)
    epi_p2 = epi[epi["PLACE"] == p_code]
    merged_e = epi_p2.merge(hea, how="inner", on="CHILD")
    merged_e = merged_e.dropna(subset=["DECOM", "DEC", "DOB"])
    error_mask = merged_e["DECOM"] < (
        merged_e["DOB"] + pd.offsets.DateOffset(years=y_gap)
    )
    return {"Episodes": merged_e["index"][error_mask].unique().tolist()}


def dec_after_decom(dfs, p_code, y_gap):
    epi = dfs["Episodes"]
    hea = dfs["Header"]

    hea["DOB"] = pd.to_datetime(hea["DOB"], format="%d/%m/%Y", errors="coerce")
    epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
    epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")

    epi.reset_index(inplace=True)
    epi_p2 = epi[epi["PLACE"] == p_code]
    merged_e = epi_p2.merge(hea, how="inner", on="CHILD")
    merged_e = merged_e.dropna(subset=["DECOM", "DEC", "DOB"])
    error_mask = merged_e["DEC"] > (
        merged_e["DECOM"] + pd.offsets.DateOffset(days=y_gap)
    )
    return {"Episodes": merged_e["index"][error_mask].unique().tolist()}


def field_different_from_previous(dfs, field):
    if "Episodes" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        collection_year = dfs["metadata"]["collectionYear"]

        epi = dfs["Episodes"]
        epi_last = dfs["Episodes_last"]

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi_last["DECOM"] = pd.to_datetime(
            epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        epi_last["DEC"] = pd.to_datetime(
            epi_last["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        epi.reset_index(inplace=True)

        first_ep_inds = epi.groupby(["CHILD"])["DECOM"].idxmin(skipna=True)
        min_decom = epi.loc[first_ep_inds, :]

        last_ep_inds = epi_last.groupby(["CHILD"])["DECOM"].idxmax(skipna=True)
        max_last_decom = epi_last.loc[last_ep_inds, :]

        merged_co = min_decom.merge(
            max_last_decom, how="inner", on=["CHILD"], suffixes=["", "_PRE"]
        )

        this_one = field
        pre_one = this_one + "_PRE"

        err_mask = merged_co[this_one].astype(str) != merged_co[pre_one].astype(str)
        err_mask = err_mask & merged_co["DEC_PRE"].isna()

        if (field == "PLACE") & (collection_year == "2024"):
            err_mask = err_mask & ~(
                (merged_co[this_one].astype(str) == "K3")
                & (merged_co[pre_one].astype(str) == "H5")
            )

        err_list = merged_co["index"][err_mask].unique().tolist()
        err_list.sort()
        return {"Episodes": err_list}


def valid_date(dte):
    try:
        lst = dte.split("/")
    except AttributeError:
        return pd.NaT
    # Preceding block checks for the scenario where the value passed in is nan/naT

    # date should have three elements
    if len(lst) != 3:
        return pd.NaT

    z_list = ["ZZ", "ZZ", "ZZZZ"]
    # We set the date to the latest possible value to avoid false positives
    offset_list = [
        pd.DateOffset(months=1, days=-1),
        pd.DateOffset(years=1, days=-1),
        None,
    ]
    # that is, go to the next month/year and take the day before that
    date_bits = []

    for i, zeds, offset in zip(lst, z_list, offset_list):
        if i == "ZZ":
            i = "01"
            offset_to_use = offset
        elif i == "ZZZZ":
            i = "2000"
        date_bits.append(i)

    as_datetime = pd.to_datetime(
        "/".join(date_bits), format="%d/%m/%Y", errors="coerce"
    )
    try:
        as_datetime += offset_to_use
    except (NameError, TypeError):  # offset_to_use only defined if needed
        pass
    return as_datetime



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="523",
    message="Date of decision that the child should be placed for adoption should be the same date as the decision that adoption is in the best interest (date should be placed).",
    affected_fields=["DATE_PLACED", "DATE_INT"],
    tables=["PlacedAdoption", "AD1"],
)
def validate(dfs):
    if ("AD1" not in dfs) or ("PlacedAdoption" not in dfs):
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        ad1 = dfs["AD1"]
        # keep initial index values to be reused for locating errors later on.
        placed_adoption.reset_index(inplace=True)
        ad1.reset_index(inplace=True)

        # convert to datetime to enable comparison
        placed_adoption["DATE_PLACED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        ad1["DATE_INT"] = pd.to_datetime(
            ad1["DATE_INT"], format="%d/%m/%Y", errors="coerce"
        )

        # drop rows where either of the required values have not been filled.
        placed_adoption = placed_adoption[placed_adoption["DATE_PLACED"].notna()]
        ad1 = ad1[ad1["DATE_INT"].notna()]

        # bring corresponding values together from both dataframes
        merged_df = placed_adoption.merge(
            ad1, on=["CHILD"], how="inner", suffixes=["_PA", "_AD"]
        )
        # find error values
        different_dates = merged_df["DATE_INT"] != merged_df["DATE_PLACED"]
        # map error locations to corresponding indices
        pa_error_locations = merged_df.loc[different_dates, "index_PA"]
        ad1_error_locations = merged_df.loc[different_dates, "index_AD"]

        return {
            "PlacedAdoption": pa_error_locations.to_list(),
            "AD1": ad1_error_locations.to_list(),
        }


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        {
            "DATE_PLACED": [
                "08/03/2020",
                "22/06/2020",
                "13/10/2022",
                "24/10/2021",
                pd.NA,
                pd.NA,
                "£1st/Februart",
                "11/11/2020",
            ],
            "CHILD": [1, 2, 3, 4, 5, 6, 7, 10],
        }
    )
    fake_ad1 = pd.DataFrame(
        {
            "DATE_INT": [
                "08/03/2020",
                "22/07/2020",
                "13/10/2021",
                "22/06/2020",
                "26/05/2018",
                pd.NA,
                "01/10/2020",
                "01/10/2020",
            ],
            "CHILD": [1, 2, 3, 4, 5, 6, 7, 333],
        }
    )

    fake_dfs = {"AD1": fake_ad1, "PlacedAdoption": fake_placed_adoption}

    # get the error function and run it on the fake data

    result = validate(fake_dfs)
    assert result == {"AD1": [1, 2, 3], "PlacedAdoption": [1, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="358",
    message="Child with this legal status should not be under 10.",
    affected_fields=["DECOM", "DOB", "LS"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        header = dfs["Header"]
        code_list = ["J1", "J2", "J3"]

        # convert dates to datetime format
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        # prepare to merge
        episodes.reset_index(inplace=True)
        header.reset_index(inplace=True)
        merged = episodes.merge(
            header, on="CHILD", how="left", suffixes=["_eps", "_er"]
        )

        # Where <LS> = ‘J1’ or ‘J2’ or ‘J3’ then <DOB> should <= to 10 years prior to <DECOM>
        mask = merged["LS"].isin(code_list) & (
            merged["DOB"] + pd.offsets.DateOffset(years=10) >= merged["DECOM"]
        )
        # That is, raise error if DECOM <= DOB + 10years

        # error locations
        header_error_locs = merged.loc[mask, "index_er"]
        episode_error_locs = merged.loc[mask, "index_eps"]
        # one to many join implies use .unique on the 'one'
        return {
            "Episodes": episode_error_locs.tolist(),
            "Header": header_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {"CHILD": 101, "DECOM": "01/01/2009", "LS": "J3"},  # 0 Fail
            {"CHILD": 102, "DECOM": "01/01/2001", "LS": "X"},  # 1
            {"CHILD": 103, "DECOM": "20/12/2001", "LS": "L2"},  # 2
            {"CHILD": 104, "DECOM": "03/01/2018", "LS": "J2"},  # 3 Pass
            {"CHILD": 105, "DECOM": "03/04/2008", "LS": "J1"},  # 4 Fail
            {"CHILD": 106, "DECOM": "01/01/2002", "LS": "L2"},  # 5
            {"CHILD": 107, "DECOM": "10/01/2002", "LS": "X"},  # 6
            {"CHILD": 108, "DECOM": "11/02/2012", "LS": "J1"},  # 7 Pass
            {"CHILD": 109, "DECOM": "25/01/2002", "LS": "L2"},  # 8
            {"CHILD": 110, "DECOM": "25/01/2002", "LS": "L2"},  # 9
            {"CHILD": 111, "DECOM": pd.NA, "LS": "J2"},  # 10 ignored by dropna
            {"CHILD": 112, "DECOM": "25/01/2002", "LS": "J3"},  # 11 ignored by dropna
        ]
    )
    fake_data_header = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DOB": "01/01/2000",
            },  # 0
            {
                "CHILD": 102,
                "DOB": "01/01/2001",
            },  # 1
            {
                "CHILD": 103,
                "DOB": "20/12/2001",
            },  # 2
            {
                "CHILD": 104,
                "DOB": "01/01/2000",
            },  # 3
            {
                "CHILD": 105,
                "DOB": "03/01/2000",
            },  # 4
            {
                "CHILD": 106,
                "DOB": "01/01/2002",
            },  # 5
            {
                "CHILD": 107,
                "DOB": "10/01/2002",
            },  # 6
            {
                "CHILD": 108,
                "DOB": "11/01/2002",
            },  # 7
            {
                "CHILD": 109,
                "DOB": "25/01/2002",
            },  # 8
            {
                "CHILD": 110,
                "DOB": "25/01/2002",
            },  # 9
            {
                "CHILD": 111,
                "DOB": "25/01/2002",
            },  # 10
            {
                "CHILD": 112,
                "DOB": pd.NA,
            },  # 11
        ]
    )
    fake_dfs = {"Episodes": fake_data_eps, "Header": fake_data_header}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 4], "Header": [0, 4]}



import pandas as pd
import numpy as np

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1016",
    message="Care leaver information is not required for 17- or 18-year olds who cease to be looked after, after their birthday.",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "OC3"],
)
def validate(dfs):
    # If <DOB> < 19 and >= to 17 years prior to <COLLECTION_END_DATE>
    # and
    # current episode <DEC> is present and after the birthdate then
    # <IN_TOUCH>, <ACTIV> and <ACCOM> should not be provided
    if ("OC3" not in dfs) | ("Episodes" not in dfs):
        return {}
    else:
        OC3 = dfs["OC3"]
        episodes = dfs["Episodes"]
        collection_end = dfs["metadata"]["collection_end"]

        OC3["DOB"] = pd.to_datetime(OC3["DOB"], format="%d/%m/%Y")
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        OC3 = OC3.reset_index()

        OC3["AGE_AT_CE"] = collection_end - OC3["DOB"]

        df_merged = OC3.merge(episodes, on="CHILD", how="inner")

        # Fails rows where the child is between 19 and 17 at the end of a return,
        # and their DEC is after their 17th birthday
        df_merged["age at dec"] = (
            df_merged["DEC"] - df_merged["DOB"]
        ) / np.timedelta64(1, "Y")

        condition_17 = df_merged["DEC"] >= (
            df_merged["DOB"] + pd.offsets.DateOffset(years=17)
        )
        condition_18 = df_merged["DEC"] >= (
            df_merged["DOB"] + pd.offsets.DateOffset(years=18)
        )

        df_errors = df_merged[
            (
                (df_merged["AGE_AT_CE"] >= np.timedelta64(17, "Y"))
                & (df_merged["AGE_AT_CE"] < np.timedelta64(18, "Y"))
                & condition_17
            )
            | (
                (df_merged["AGE_AT_CE"] >= np.timedelta64(18, "Y"))
                & (df_merged["AGE_AT_CE"] < np.timedelta64(19, "Y"))
                & condition_18
            )
        ]

        df_errors = df_errors[
            df_errors["IN_TOUCH"].notna()
            | df_errors["ACTIV"].notna()
            | df_errors["ACCOM"].notna()
        ]

        return {"OC3": df_errors["index"].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "DOB": "31/03/2002",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child2",
                "DOB": "31/03/2002",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child3",
                "DOB": "31/03/2002",
                "IN_TOUCH": pd.NA,
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },
            {
                "CHILD": "child4",
                "DOB": "31/03/2003",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child5",
                "DOB": "31/03/2003",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child6",
                "DOB": "31/03/2002",
                "IN_TOUCH": pd.NA,
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },
            {
                "CHILD": "child7",
                "DOB": "01/01/2002",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },  # Not in EPI file, IGNORE
            {
                "CHILD": "child8",
                "DOB": "31/03/2004",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child9",
                "DOB": "31/03/2004",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child10",
                "DOB": "31/03/2002",
                "IN_TOUCH": pd.NA,
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },
            {
                "CHILD": "child11",
                "DOB": "31/03/2002",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
            {
                "CHILD": "child12",
                "DOB": "31/03/2003",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "Y",
            },
        ]
    )

    fake_epi = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "DEC": "31/03/2020",
            },  # 18 within collection year and BDAY on DEC,  FAIL
            {
                "CHILD": "child2",
                "DEC": "30/03/2020",
            },  # 18 within collection year and BDAY after DEC, PASS
            {
                "CHILD": "child3",
                "DEC": "31/03/2020",
            },  # 18 within collection year and BDAY on DEC with no IN TOUCH, PASS
            {
                "CHILD": "child4",
                "DEC": "31/03/2020",
            },  # 17 within collection year and BDAY on DEC, FAIL
            {
                "CHILD": "child5",
                "DEC": "01/01/2020",
            },  # 17 within collection year and BDAY after DEC, PASS
            {
                "CHILD": "child6",
                "DEC": "31/03/2020",
            },  # 17 within collection year and BDAY on DEC with no IN TOUCH, PASS
            {
                "CHILD": "child8",
                "DEC": "31/03/2020",
            },  # 16 within collection year and BDAY on DEC, IGNORE
            {
                "CHILD": "child9",
                "DEC": "30/03/2020",
            },  # 16 within collection year and BDAY after DEC, IGNORE
            {
                "CHILD": "child10",
                "DEC": "30/03/2020",
            },  # 18 within collection year and BDAY before DEC but no IN TOUCH, IGNORE
            {
                "CHILD": "child11",
                "DEC": pd.NA,
            },  # 18 within collection year but no DEC, IGNORE
            {
                "CHILD": "child12",
                "DEC": pd.NA,
            },  # 17 within collection year but no DEC, IGNORE
        ]
    )

    metadata = {"collection_end": pd.to_datetime("31/03/2020", format="%d/%m/%Y")}

    fake_dfs = {
        "Episodes": fake_epi,
        "OC3": fake_data,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"OC3": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="504",
    message="The category of need code differs from that reported at start of current period of being looked after",
    affected_fields=["CIN"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")

        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)
        epi["LAG_INDEX"] = epi["level_0"].shift(1)

        merge_epi = epi.merge(
            epi,
            how="inner",
            left_on="LAG_INDEX",
            right_on="level_0",
            suffixes=["", "_PRE"],
        )
        merge_epi = merge_epi[merge_epi["CHILD"] == merge_epi["CHILD_PRE"]]
        merge_epi = merge_epi[
            (merge_epi["REC_PRE"] == "X1")
            & (merge_epi["DEC_PRE"] == merge_epi["DECOM"])
        ]
        error_cohort = merge_epi[merge_epi["CIN"] != merge_epi["CIN_PRE"]]
        error_list = error_cohort["index"].unique().tolist()
        error_list.sort()
        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/06/2020",
                "DEC": "04/06/2020",
                "REC": "X1",
                "CIN": "N1",
            },  # 0
            {
                "CHILD": "111",
                "DECOM": "04/06/2020",
                "DEC": "06/06/2020",
                "REC": "X1",
                "CIN": "N2",
            },  # 1   Fails
            {
                "CHILD": "111",
                "DECOM": "06/06/2020",
                "DEC": pd.NA,
                "REC": "X1",
                "CIN": "N3",
            },  # 2   Fails
            {
                "CHILD": "111",
                "DECOM": "08/06/2020",
                "DEC": "09/06/2020",
                "REC": "X1",
                "CIN": "N2",
            },  # 3
            {
                "CHILD": "222",
                "DECOM": "10/06/2020",
                "DEC": "11/06/2020",
                "REC": "X1",
                "CIN": "N3",
            },  # 4
            {
                "CHILD": "333",
                "DECOM": "04/06/2020",
                "DEC": "07/06/2020",
                "REC": "X1",
                "CIN": "N4",
            },  # 5
            {
                "CHILD": "333",
                "DECOM": "07/06/2020",
                "DEC": pd.NA,
                "REC": "X1",
                "CIN": "N1",
            },  # 6   Fails
            {
                "CHILD": "444",
                "DECOM": "08/06/2020",
                "DEC": "09/06/2020",
                "REC": "X1",
                "CIN": "N2",
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "08/06/2020",
                "DEC": "10/06/2020",
                "REC": "X1",
                "CIN": "N3",
            },  # 8
            {
                "CHILD": "444",
                "DECOM": "10/06/2020",
                "DEC": pd.NA,
                "REC": "X1",
                "CIN": "N4",
            },  # 9   Fails
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 6, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="232",
    message="Sex of the child/young person has been reported as unknown.",
    affected_fields=["SEX"],
    tables=["Header"],
)
def validate(dfs):
    # If <UASC_STATUS> = 1 then <DOB> should be > 11 years prior to <COLLECTION_END_DATE>
    if "Header" not in dfs:
        return {}
    else:
        df = dfs["Header"]

        error_df = df[df["SEX"].str.upper() == "U"]

        errors = error_df.index

        return {"Header": errors.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"SEX": "U"},  # sex U, fails
            {"SEX": "u"},  # sex u fails
            {"SEX": "m"},
            {"SEX": 0},
        ]
    )  # over 11, uasc, passes

    fake_metadata = {"collection_end": "31/03/2025"}

    fake_dfs = {"Header": fake_data, "metadata": fake_metadata}

    result = validate(fake_dfs)

    assert result == {"Header": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="517",
    message="If reporting legal status of adopters is L3 then the genders of adopters should be coded as MF. MF = the adopting couple are male and female.",
    affected_fields=["LS_ADOPTR", "SEX_ADOPTR"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}

    else:
        AD1 = dfs["AD1"]

        error_mask = AD1["LS_ADOPTR"].eq("L3") & ~AD1["SEX_ADOPTR"].isin(["MF"])

        error_locations = AD1.index[error_mask]

        return {"AD1": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS_ADOPTR": ["L3", "L1", pd.NA, "L3", "L3", "L4", "L3"],
            "SEX_ADOPTR": ["MF", "MF", "MM", "M1", "FF", "F1", "xxxxx"],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {"AD1": [3, 4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="149",
    message="Reason episode ceased code is not valid. ",
    affected_fields=["REC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    episodes = dfs["Episodes"]
    code_list = [
        "E11",
        "E12",
        "E2",
        "E3",
        "E4A",
        "E4B",
        "E13",
        "E41",
        "E45",
        "E46",
        "E47",
        "E48",
        "E5",
        "E6",
        "E7",
        "E8",
        "E9",
        "E14",
        "E15",
        "E16",
        "E17",
        "X1",
    ]

    mask = episodes["REC"].isin(code_list) | episodes["REC"].isna()

    validation_error_mask = ~mask
    validation_error_locations = episodes.index[validation_error_mask]

    return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REC": ["E4A", "E4b", "X", "", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="134",
    message="Data on adoption should not be entered for the OC3 cohort.",
    affected_fields=[
        "IN_TOUCH",
        "ACTIV",
        "ACCOM",
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ],
    tables=["OC3", "AD1"],
)
def validate(dfs):
    if "OC3" not in dfs or "AD1" not in dfs:
        return {}
    else:
        oc3 = dfs["OC3"]
        ad1 = dfs["AD1"]
        ad1["ad1_index"] = ad1.index

        all_data = ad1.merge(oc3, how="left", on="CHILD")

        na_oc3_data = (
            all_data["IN_TOUCH"].isna()
            & all_data["ACTIV"].isna()
            & all_data["ACCOM"].isna()
        )

        na_ad1_data = (
            all_data["DATE_INT"].isna()
            & all_data["DATE_MATCH"].isna()
            & all_data["FOSTER_CARE"].isna()
            & all_data["NB_ADOPTR"].isna()
            & all_data["SEX_ADOPTR"].isna()
            & all_data["LS_ADOPTR"].isna()
        )

        validation_error = ~na_oc3_data & ~na_ad1_data
        validation_error_locations = all_data.loc[
            validation_error, "ad1_index"
        ].unique()

        return {"AD1": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E"],
            "IN_TOUCH": [pd.NA, "XXX", pd.NA, pd.NA, pd.NA],
            "ACTIV": [pd.NA, pd.NA, "XXX", pd.NA, pd.NA],
            "ACCOM": [pd.NA, pd.NA, pd.NA, "XXX", pd.NA],
        }
    )
    fake_data_ad1 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E"],
            "DATE_INT": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "DATE_MATCH": [pd.NA, "XXX", "XXX", pd.NA, "XXX"],
            "FOSTER_CARE": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "NB_ADOPTR": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "SEX_ADOPTR": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "LS_ADOPTR": [pd.NA, pd.NA, "XXX", "XXX", "XXX"],
        }
    )

    fake_dfs = {"OC3": fake_data_oc3, "AD1": fake_data_ad1}

    result = validate(fake_dfs)

    assert result == {"AD1": [1, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="580",
    message="Child is missing when cease being looked after but reason episode ceased not ‘E8’.",
    affected_fields=["REC"],
    tables=["Episodes", "MIssing"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Missing" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        mis = dfs["Missing"]
        mis["MIS_END"] = pd.to_datetime(
            mis["MIS_END"], format="%d/%m/%Y", errors="coerce"
        )
        mis["DOB"] = pd.to_datetime(mis["DOB"], format="%d/%m/%Y", errors="coerce")
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")

        epi.reset_index(inplace=True)
        mis["BD18"] = mis["DOB"] + pd.DateOffset(years=18)

        m_coh = mis.merge(epi, how="inner", on="CHILD")
        m_coh = m_coh.query("(BD18 == MIS_END) & (MIS_END == DEC) & DEC.notnull()")
        err_list = m_coh.query("REC != 'E8'")["index"].unique().tolist()
        err_list.sort()
        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        [
            {"CHILD": "111", "DEC": "01/06/2020", "REC": "E7"},  # 0 fails
            {"CHILD": "123", "DEC": "08/06/2020", "REC": "E7"},  # 1 fails
            {"CHILD": "222", "DEC": "05/06/2020", "REC": "E8"},  # 2
            {"CHILD": "333", "DEC": "06/06/2020", "REC": "E5"},  # 3 fails
            {"CHILD": "444", "DEC": "07/06/2020", "REC": "E8"},  # 4
            {"CHILD": "555", "DEC": "19/06/2020", "REC": "E4"},  # 5 fails
            {"CHILD": "777", "DEC": "19/06/2020", "REC": "E4"},  # 6 fails
        ]
    )

    fake_mis = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "MISSING": "M",
                "MIS_END": "01/06/2020",
                "DOB": "01/06/2002",
            },  #
            {
                "CHILD": "123",
                "MISSING": "M",
                "MIS_END": "08/06/2020",
                "DOB": "08/06/2002",
            },  #
            {
                "CHILD": "222",
                "MISSING": "M",
                "MIS_END": "05/06/2020",
                "DOB": "05/06/2002",
            },  #
            {
                "CHILD": "333",
                "MISSING": "M",
                "MIS_END": "06/06/2020",
                "DOB": "06/06/2002",
            },  #
            {
                "CHILD": "444",
                "MISSING": "M",
                "MIS_END": "08/06/2020",
                "DOB": "08/06/2002",
            },  #
            {
                "CHILD": "444",
                "MISSING": "M",
                "MIS_END": "09/06/2020",
                "DOB": "09/06/2002",
            },  #
            {
                "CHILD": "555",
                "MISSING": "M",
                "MIS_END": "19/06/2020",
                "DOB": "19/06/2002",
            },  #
            {
                "CHILD": "777",
                "MISSING": "A",
                "MIS_END": "19/06/2020",
                "DOB": "19/06/2002",
            },  #
        ]
    )

    fake_dfs = {"Episodes": fake_epi, "Missing": fake_mis}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 1, 3, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="228",
    message="Ofsted Unique reference number (URN) is not valid for the episode end date "
    "[NOTE: may give false positives on open episodes at providers who close during the year]",
    affected_fields=["URN", "DEC"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("provider_info" not in dfs["metadata"]):
        return {}
    else:
        episodes = dfs["Episodes"]
        provider_info = dfs["metadata"]["provider_info"]
        collection_end = dfs["metadata"]["collection_end"]

        # convert date fields from strings to datetime format. NB. REG_END is in datetime format already.
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        # merge
        episodes["index_eps"] = episodes.index
        episodes = episodes[episodes["URN"].notna() & (episodes["URN"] != "XXXXXXX")]
        provider_info = provider_info[provider_info["REG_END"].notna()]

        merged = episodes.merge(provider_info, on="URN", how="inner")
        # If <URN> provided and not = 'XXXXXXX', and Ofsted URN <REG_END> not NULL then <DEC> if provided
        # must be <= Ofsted <REG_END>OR if not provided then<COLLECTION_END_DATE>must be<= <REG_END>.
        # Note: For open episodes (those without an end date) a check should be made to ensure that the Ofsted
        # URN was still open at the 31 March of the current year.
        mask = (merged["DEC"].notna() & (merged["DEC"] > merged["REG_END"])) | (
            merged["DEC"].isna() & (collection_end > merged["REG_END"])
        )

        eps_error_locations = merged.loc[mask, "index_eps"].sort_values().to_list()
        return {"Episodes": eps_error_locations}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "DEC": pd.NA,
                "URN": 1,
            },  # 0 pass REG_END is after March 31st of collection year
            {
                "CHILD": "1111",
                "DEC": "01/02/2015",
                "URN": pd.NA,
            },  # 1 ignore: URN not provided
            {
                "CHILD": "1111",
                "DEC": pd.NA,
                "URN": 3,
            },  # 2 fail REG_END is before March 31st of collection year
            {
                "CHILD": "2222",
                "DEC": "01/01/2010",
                "URN": "XXXXXXX",
            },  # 3 ignore: URN
            {
                "CHILD": "3333",
                "DEC": "01/01/2010",
                "URN": 2,
            },  # 4 pass
            {
                "CHILD": "3333",
                "DEC": "25/12/2015",
                "URN": 2,
            },  # 5 fail DEC after REG_END
            {
                "CHILD": "4444",
                "DEC": "25/12/2016",
                "URN": 1,
            },  # 6 fail. DEC after REG_END
            {
                "CHILD": "5555",
                "DEC": "01/01/2010",
                "URN": 4,
            },  # 7 ignore: REG_END is null
            {
                "CHILD": "5555",
                "DEC": "25/12/2015",
                "URN": 1,
            },  # 8 pass DEC equals REG_END
        ]
    )
    fake_provider_info = pd.DataFrame(
        [
            {
                "URN": 1,
                "REG_END": "25/12/2015",
            },  # 0
            {
                "URN": 2,
                "REG_END": "21/02/2014",
            },  # 1
            {
                "URN": 3,
                "REG_END": "01/02/2015",
            },  # 2
            {
                "URN": 4,
                "REG_END": pd.NA,
            },  # 3
        ]
    )
    fake_provider_info["REG_END"] = pd.to_datetime(
        fake_provider_info["REG_END"], format="%d/%m/%Y", errors="raise"
    )
    metadata = {
        "collection_start": "01/04/2014",
        "collection_end": "31/03/2015",
        "provider_info": fake_provider_info,
    }

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="525",
    message="A child for whom the decision to be placed for adoption has been reversed cannot be adopted during the year.",
    affected_fields=[
        "DATE_PLACED_CEASED",
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ],
    tables=["PlacedAdoption", "AD1"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs or "AD1" not in dfs:
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        ad1 = dfs["AD1"]
        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        ad1.reset_index(inplace=True)

        merged = placed_adoption.merge(
            ad1, on="CHILD", how="left", suffixes=["_placed", "_ad1"]
        )
        # If <DATE_PLACED_CEASED> not Null, then <DATE_INT>; <DATE_MATCH>; <FOSTER_CARE>; <NB_ADOPTR>; <SEX_ADOPTR>; and <LS_ADOPTR> should not be provided
        mask = merged["DATE_PLACED_CEASED"].notna() & (
            merged["DATE_INT"].notna()
            | merged["DATE_MATCH"].notna()
            | merged["FOSTER_CARE"].notna()
            | merged["NB_ADOPTR"].notna()
            | merged["SEX_ADOPTR"].notna()
            | merged["LS_ADOPTR"].notna()
        )
        # error locations
        pa_error_locs = merged.loc[mask, "index_placed"]
        ad_error_locs = merged.loc[mask, "index_ad1"]
        # return result
        return {
            "PlacedAdoption": pa_error_locs.tolist(),
            "AD1": ad_error_locs.tolist(),
        }


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        {
            "DATE_PLACED_CEASED": [
                "08/03/2020",
                "22/06/2020",
                "13/10/2022",
                "24/10/2021",
                pd.NA,
            ],
            "CHILD": ["104", "105", "107", "108", "109"],
        }
    )
    fake_data_ad1 = pd.DataFrame(
        {
            "CHILD": ["104", "105", "107", "108", "109"],
            "DATE_INT": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "DATE_MATCH": [pd.NA, "XXX", "XXX", pd.NA, "XXX"],
            "FOSTER_CARE": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "NB_ADOPTR": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "SEX_ADOPTR": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "LS_ADOPTR": [pd.NA, pd.NA, "XXX", "XXX", "XXX"],
        }
    )

    fake_dfs = {"PlacedAdoption": fake_placed_adoption, "AD1": fake_data_ad1}

    result = validate(fake_dfs)
    assert result == {"PlacedAdoption": [1, 2, 3], "AD1": [1, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="203",
    message="Date of birth disagrees with the date of birth already recorded for this child.",
    affected_fields=["DOB"],
    tables=["Header", "Header_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        header_last = dfs["Header_last"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        header_last["DOB"] = pd.to_datetime(
            header_last["DOB"], format="%d/%m/%Y", errors="coerce"
        )

        header_merged = (
            header.reset_index()
            .merge(
                header_last,
                how="left",
                on=["CHILD"],
                suffixes=("", "_last"),
                indicator=True,
            )
            .set_index("index")
        )

        in_both_years = header_merged["_merge"] == "both"
        dob_is_different = header_merged["DOB"].astype(str) != header_merged[
            "DOB_last"
        ].astype(str)

        error_mask = in_both_years & dob_is_different

        error_locations = header.index[error_mask]

        return {"Header": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108", "109", "110"],
            "DOB": [
                "16/03/2020",
                "23/09/2016",
                "31/12/19",
                "31/02/2018",
                pd.NA,
                "10/08/2014",
                pd.NA,
                "20/01/2017",
                "31/06/2020",
            ],
        }
    )

    fake_data_prev = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "107", "108", "109", "110"],
            "DOB": [
                "16/03/2020",
                "22/09/2016",
                "31/12/2019",
                "31/02/2018",
                pd.NA,
                "11/11/2021",
                "04/06/2017",
                pd.NA,
                "30/06/2020",
            ],
        }
    )

    fake_dfs = {"Header": fake_data, "Header_last": fake_data_prev}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 2, 6, 7, 8]}



import pandas as pd

from lac_validator.fixtures import fake_INT_file, fake_INT_header
from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="INT17",
    message="Internal Check: DOB in Reviews is different to DOB in Header.",
    affected_fields=["DOB"],
    tables=["Header", "Reviews"],
)
def validate(dfs):
    if "Header" not in dfs or "Reviews" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        file = dfs["Reviews"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        file["DOB"] = pd.to_datetime(file["DOB"], format="%d/%m/%Y", errors="coerce")

        file["index_file"] = file.index

        merged = header.merge(
            file[["CHILD", "DOB", "index_file"]],
            on="CHILD",
            indicator=True,
            how="right",
            suffixes=["_header", "_file"],
        )

        mask = (
            (merged["DOB_header"] != merged["DOB_file"])
            & (merged["DOB_header"].notna() & merged["DOB_file"].notna())
            & (merged["_merge"] != "right_only")
        )
        eps_error_locations = merged.loc[mask, "index_file"]
        return {"Reviews": eps_error_locations.unique().tolist()}


def test_validate():
    fake_dfs = {"Header": fake_INT_header.copy(), "Reviews": fake_INT_file.copy()}
    result = validate(fake_dfs)
    assert result == {"Reviews": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW03STG1",
    message="For each social worker episode, information should be complete.",
    affected_fields=["SW_ID", "SW_DECOM", "SW_REASON"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        SWE = dfs["SWEpisodes"]

        error_df = SWE[
            SWE["SW_ID"].isna() | SWE["SW_DECOM"].isna() | SWE["SW_REASON"].isna()
        ]

        error_rows = error_df.index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_swe = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "SW_ID": pd.NA,
                "SW_DECOM": "xx",
                "SW_DEC": "XX",
                "SW_REASON": "xx",
            },
            {
                "CHILD": "child2",
                "SW_ID": "xx",
                "SW_DECOM": pd.NA,
                "SW_DEC": "XX",
                "SW_REASON": "xx",
            },
            {
                "CHILD": "child3",
                "SW_ID": "xx",
                "SW_DECOM": "xx",
                "SW_DEC": "XX",
                "SW_REASON": pd.NA,
            },
            {
                "CHILD": "child4",
                "SW_ID": "xx",
                "SW_DECOM": "xx",
                "SW_DEC": pd.NA,
                "SW_REASON": "xx",
            },
            {
                "CHILD": "child5",
                "SW_ID": "xx",
                "SW_DECOM": "xx",
                "SW_DEC": "XX",
                "SW_REASON": "xx",
            },
        ]
    )

    fake_dfs = {"SWEpisodes": fake_swe}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [0, 1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="344",
    message="The record shows the young person has died or returned home to live with parent(s) or someone with parental responsibility for a continuous period of 6 months or more, but activity and/or accommodation on leaving care have been completed.",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}
    else:
        oc3 = dfs["OC3"]
        # If <IN_TOUCH> = 'DIED' or 'RHOM' then <ACTIV> and <ACCOM> should not be provided (0 for this)
        mask = ((oc3["IN_TOUCH"] == "DIED") | (oc3["IN_TOUCH"] == "RHOM")) & (
            (oc3["ACTIV"].astype(str) != "0") | (oc3["ACCOM"].astype(str) != "0")
        )
        error_locations = oc3.index[mask]
        return {"OC3": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E", "F"],
            "IN_TOUCH": ["DIED", "Yes", "RHOM", pd.NA, "DIED", "RHOM"],
            "ACTIV": ["0", pd.NA, "XXX", pd.NA, "XXX", pd.NA],
            "ACCOM": [0, pd.NA, 0, "XXX", "XXX", 0],
        }
    )
    fake_dfs = {"OC3": fake_data_oc3}

    result = validate(fake_dfs)
    assert result == {"OC3": [2, 4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1011",
    message="This child is recorded as having his/her care transferred to another local authority for the final episode and therefore should not have the care leaver information completed.",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "OC3"],
)
def validate(dfs):
    if "OC3" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        oc3 = dfs["OC3"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")

        # If final <REC> = 'E3' then <IN_TOUCH>; <ACTIV> and <ACCOM> should not be provided
        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        grouped_decom_by_child = epi.groupby(["CHILD"])["DECOM"].idxmax(skipna=True)
        max_decom_only = epi.loc[epi.index.isin(grouped_decom_by_child), :]
        E3_is_last = max_decom_only[max_decom_only["REC"] == "E3"]

        oc3.reset_index(inplace=True)
        cohort_to_check = oc3.merge(E3_is_last, on="CHILD", how="inner")
        error_mask = cohort_to_check[["IN_TOUCH", "ACTIV", "ACCOM"]].notna().any(axis=1)

        error_list = cohort_to_check["index"][error_mask].to_list()
        error_list = list(set(error_list))
        error_list.sort()

        return {"OC3": error_list}


def test_validate():
    import pandas as pd

    fake_data_epi = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "01/06/2020", "REC": "E3"},  # 0
            {"CHILD": "111", "DECOM": "05/06/2020", "REC": "E11"},  # 1
            {"CHILD": "111", "DECOM": "06/06/2020", "REC": "E3"},  # 2  Max E3
            {"CHILD": "123", "DECOM": "08/06/2020", "REC": "X1"},  # 3
            {"CHILD": "222", "DECOM": "05/06/2020", "REC": "E3"},  # 4   Max E3
            {"CHILD": "333", "DECOM": "06/06/2020", "REC": "X1"},  # 5
            {"CHILD": "333", "DECOM": "07/06/2020", "REC": "E3"},  # 6  Max E3
            {"CHILD": "444", "DECOM": "08/06/2020", "REC": "E3"},  # 7
            {"CHILD": "444", "DECOM": "09/06/2020", "REC": "E3"},  # 8
            {"CHILD": "444", "DECOM": "15/06/2020", "REC": "X1"},  # 9
            {"CHILD": "555", "DECOM": "15/06/2020", "REC": "E3"},  # 10  Max E3
            {"CHILD": "666", "DECOM": "15/06/2020", "REC": pd.NA},  # 11
        ]
    )
    fake_data_oc3 = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "IN_TOUCH": "Whatever",
                "ACTIV": pd.NA,
                "ACCOM": "Whatever",
            },  # 0 Fails
            {
                "CHILD": "222",
                "IN_TOUCH": pd.NA,
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },  # 1 All null values so will pass
            {
                "CHILD": "333",
                "IN_TOUCH": "Whatever",
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },  # 2 Fails
            {
                "CHILD": "777",
                "IN_TOUCH": "Whatever",
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },  # 3
            {
                "CHILD": "555",
                "IN_TOUCH": "Whatever",
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },  # 4 Fails
        ]
    )
    fake_dfs = {"OC3": fake_data_oc3, "Episodes": fake_data_epi}

    result = validate(fake_dfs)

    assert result == {"OC3": [0, 2, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="527",
    message="A placement provider code of PR1 cannot be associated with placements P1, R2 or R5.",
    affected_fields=["PLACE", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        error_mask = (epi["PLACE"].isin(["P1", "R2", "R5"])) & (
            epi["PLACE_PROVIDER"] == "PR1"
        )
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["P1", "A3", "K1", "R2", "P1", "X1"],
            "PLACE_PROVIDER": ["PR1", "PR3", "PR4", "PR1", "PR1", "PR1"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW07STG2",
    message="Social worker episode end date should be on or after the start of the social worker reporting period, and should be on or before the collection end date.",
    affected_fields=["SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]
        collection_end = dfs["metadata"]["collection_end"]
        collection_end = pd.to_datetime(collection_end, format="%d/%m/%Y")

        collection_start = dfs["metadata"]["collection_start"]
        collection_start = pd.to_datetime(collection_start, format="%d/%m/%Y")

        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")
        df = df[df["SW_DEC"].notna()].copy()

        error_rows = df[
            ~(df["SW_DEC"] >= collection_start) | ~(df["SW_DEC"] <= collection_end)
        ].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SW_DEC": ["01/04/2000", "31/03/1999", "01/01/2000", pd.NA],
        }
    )

    fake_meta = {"collection_end": "31/03/2000", "collection_start": "01/04/1999"}

    fake_dfs = {"SWEpisodes": fake_data, "metadata": fake_meta}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="530",
    message="A placement provider code of PR4 cannot be associated with placement P1.",
    affected_fields=["PLACE", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        mask = episodes["PLACE"].eq("P1") & episodes["PLACE_PROVIDER"].eq("PR4")

        validation_error_mask = mask
        validation_error_locations = episodes.index[validation_error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["P1", "A3", "K1", "P1", "P1", "P1"],
            "PLACE_PROVIDER": ["PR4", "PR3", "PR4", "PR4", "PR5", "PRO"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="380",
    message="A period of care cannot start with a temporary placement.",
    affected_fields=["PLACE", "RNE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        error_mask = (epi["PLACE"].isin(["T0", "T1", "T2", "T3", "T4"])) & (
            ~epi["RNE"].isin(["P", "B"])
        )
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "RNE": ["S", "L", "P", "B", "P", pd.NA],
            "PLACE": ["U1", "T0", "U2", "Z1", "T1", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW14STG2",
    message="Social worker episode date ceased is blank.",
    affected_fields=["SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")
        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        df["index"] = df.index

        empty_dec = df[df["SW_DEC"].isna()]

        df = df.sort_values(["CHILD", "SW_DECOM"])
        latest_episode = df.drop_duplicates("CHILD", keep="last")

        error_rows = empty_dec[~empty_dec["index"].isin(latest_episode["index"])][
            "index"
        ]

        return {"SWEpisodes": error_rows.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "1", "SW_DECOM": "01/01/2000", "SW_DEC": "01/01/2000"},
            {"CHILD": "1", "SW_DECOM": "02/01/2000", "SW_DEC": pd.NA},
            {"CHILD": "2", "SW_DECOM": "01/01/2000", "SW_DEC": pd.NA},
            {"CHILD": "2", "SW_DECOM": "02/01/2000", "SW_DEC": pd.NA},
            {"CHILD": "3", "SW_DECOM": "01/01/2000", "SW_DEC": pd.NA},
        ]
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="381",
    message="A period of care cannot end with a temporary placement.",
    affected_fields=["PLACE", "REC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        error_mask = (
            (epi["PLACE"].isin(["T0", "T1", "T2", "T3", "T4"]))
            & (epi["REC"] != "X1")
            & (epi["REC"].notna())
        )
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REC": ["X1", "PR1", "X1", pd.NA, "X1", pd.NA],
            "PLACE": ["T3", "T0", "U2", "T2", "T1", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1]}



import logging
from typing import Any, Optional

import pandas as pd
from pandas import DataFrame

from lac_validator.datastore import copy_datastore, create_datastore
from lac_validator.ingress import read_from_text
from lac_validator.rule_engine import RuleDefinition
from lac_validator.types import UploadedFile

logger = logging.getLogger(__name__)


class LacValidator:
    """
    Central location for running rules on files.
    """

    def __init__(
        self,
        metadata: dict[str, str],
        files: list[UploadedFile],
        registry: dict[str, RuleDefinition],
        selected_rules: Optional[list[str]] = None,
    ):
        self.dfs: dict[str, DataFrame] = {}
        self.dones: list[str] = []
        self.skips: list[str] = []
        self.fails: list[str] = []

        logger.info("Reading uploaded files...")
        dfs, metadata_extras = read_from_text(raw_files=files)
        self.dfs = dfs

        metadata.update(metadata_extras)
        logger.info(f'Metadata receieved: {",".join(metadata.keys())}')

        self.registry = registry
        self.metadata = metadata

        # validate
        self.validate(selected_rules)

    def get_rules_to_run(
        self,
        registry: dict[str, RuleDefinition],
        selected_rules: Optional[list[str]] = None,
    ):
        """
        Filters rules to be run based on user's selection in the frontend.
        :param dict registry: record of all existing rules in rule pack
        :param list selected_rules: array of rule codes as strings
        """
        if selected_rules:
            rules_to_run = {
                rule_code: rule
                for rule_code, rule in registry.items()
                if str(rule_code) in selected_rules
            }
            return rules_to_run
        else:
            return registry

    def validate(self, selected_rules: Optional[list[str]] = None):
        logger.info("Creating Data store...")
        data_store = create_datastore(self.dfs, self.metadata)

        rules_to_run = self.get_rules_to_run(self.registry, selected_rules)

        # this corresponds to raw_data in CINvalidationSession
        self.ds_results = copy_datastore(data_store)

        for rule_code, rule in rules_to_run.items():
            logger.info(f"Validating rule {rule_code}...")

            # get a clean copy of the files to be validated
            ds_copy = copy_datastore(data_store)

            try:
                # get the result from when the rule is run on the data.
                result: dict[str, list[Any]] = rule.func(ds_copy)
            except Exception as e:
                # document instances where the rule cannot run on the data
                logger.exception(f"Rule code {rule_code} failed to run!")
                self.fails.append(rule_code)
                continue

            if result == {}:
                # TODO replace this with behaviour that models rules skipped because of lack of data.

                # validation rules return an empty dict if the required tables are not all available.
                logger.info(f"Error code {rule.code} skipped due to missing tables")
                self.skips.append(rule.code)
            else:
                self.dones.append(rule.code)

            # map failing locations back to data files.
            for table, values in result.items():
                if len(values) > 0:
                    logger.info(f"Error code {rule.code} found {len(values)} errors")
                    nof_errors = len(values)
                    # select out only the valid values, that is remove all nans.
                    values = [
                        i for i, not_nan in zip(values, pd.notna(values)) if not_nan
                    ]
                    nof_nans = nof_errors - len(values)
                    if nof_nans != 0:
                        logger.warning(
                            f"{rule.code} returned {nof_nans} NaNs! "
                            + f"Output: {str(values)}"
                        )
                    self.ds_results[table].loc[values, f"ERR_{rule.code}"] = True


def create_issue_df(report: DataFrame, error_report: DataFrame):
    """
    creates issue_df similar to that of the CIN backend output.
    Columns should be child_id, tables_affected, columns_affected, row_id, rule_code and rule_description.

    Uses output of 903validator.report because it is closer to desired structure.

    :param df report: childID, rule_code, tables_affected and row_id will be grabbed from here
    :param df error_report: rule_description and affected_fields, per rule_code, will be grabbed from here.
    :return df full_issue_df: data to drive frontend display of issue locations.
    """
    # get rule codes from column names
    col_error_names = [c for c in report.columns if c[:4] == "ERR_"]

    # loop through rule codes and fetch rule data from error_report each time.
    issue_dfs_per_rule = []
    for col_name in col_error_names:
        rule_code = col_name[4:]

        # child, table, row where rule fails.
        rule_table = report[report[col_name] == True][
            [
                "Table",
                "RowID",
                "CHILD",
            ]
            + [col_name]
        ]
        rule_table["Code"] = rule_code

        rule_desc = error_report[error_report["Code"] == rule_code]
        # assumption. error_description has one entry per rule.
        rule_issue_df = rule_table.merge(rule_desc, on="Code")

        rule_issue_df.rename(
            columns={
                "Table": "tables_affected",
                "RowID": "row_id",
                "Code": "rule_code",
                "Description": "rule_description",
                "Fields": "columns_affected",
                "CHILD": "child_id",
            },
            inplace=True,
        )
        # remove ERR_ columns
        rule_issue_df.drop(columns=[col_name], inplace=True)
        # explode column lists into one per row.
        rule_issue_df = rule_issue_df.explode(
            column=["columns_affected"], ignore_index=True
        )
        # reorder
        rule_issue_df = rule_issue_df[
            [
                "child_id",
                "tables_affected",
                "columns_affected",
                "row_id",
                "rule_code",
                "rule_description",
            ]
        ]

        issue_dfs_per_rule.append(rule_issue_df)

    full_issue_df = pd.concat(issue_dfs_per_rule)
    full_issue_df.reset_index(drop=True, inplace=True)
    return full_issue_df



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="354",
    message="Date episode ceased must be on or before the end of the current collection year.",
    affected_fields=["DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        error_mask = epi["DECOM"] > collection_end
        error_list = epi.index[error_mask].to_list()
        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DECOM": [
                "01/01/2021",
                "19/02/2010",
                "38/04/2019",
                "01/01/2022",
                "01/04/2021",
                "01/05/2021",
                pd.NA,
                "3rd Dec 1873",
            ],
        }
    )

    metadata = {"collection_end": "01/04/2021"}

    fake_dfs = {"Episodes": fake_data, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="557",
    message="Child for whom the decision was made that they should be placed for adoption has left care "
    + "but was not adopted and information on the decision that they should no longer be placed for "
    + "adoption items has not been completed.",
    affected_fields=[
        "DATE_PLACED_CEASED",
        "REASON_PLACED_CEASED",  # PlacedAdoption
        "PLACE",
        "LS",
        "REC",
    ],  # Episodes
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    if "PlacedAdoption" not in dfs:
        return {}
    else:
        eps = dfs["Episodes"]
        placed = dfs["PlacedAdoption"]

        eps = eps.reset_index()
        placed = placed.reset_index()

        child_placed = eps["PLACE"].isin(["A3", "A4", "A5", "A6"])
        order_granted = eps["LS"].isin(["D1", "E1"])
        not_adopted = (
            ~eps["REC"].isin(["E11", "E12"]) & eps["REC"].notna() & (eps["REC"] != "X1")
        )

        placed["ceased_incomplete"] = (
            placed["DATE_PLACED_CEASED"].isna() | placed["REASON_PLACED_CEASED"].isna()
        )

        eps = eps[(child_placed | order_granted) & not_adopted]

        eps = eps.merge(
            placed, on="CHILD", how="left", suffixes=["_EP", "_PA"], indicator=True
        )

        eps = eps[(eps["_merge"] == "left_only") | eps["ceased_incomplete"]]

        EP_errors = eps["index_EP"]
        PA_errors = eps["index_PA"].dropna()

        return {
            "Episodes": EP_errors.to_list(),
            "PlacedAdoption": PA_errors.to_list(),
        }


def test_validate():
    import pandas as pd

    test_episodes = pd.DataFrame(
        [
            {
                "CHILD": "11",
                "PLACE": "A3",
                "LS": "x",
                "REC": "x",
            },  # 0: Fail! (missing vals in PA)
            {
                "CHILD": "202",
                "PLACE": "x",
                "LS": "D1",
                "REC": "x",
            },  # 1: PA filled in --ok
            {"CHILD": "3003", "PLACE": "A4", "LS": "D1", "REC": "E12"},  # 2: E12 --ok
            {"CHILD": "40004", "PLACE": "x", "LS": "E1", "REC": "E12"},  # 3: E12 --ok
            {
                "CHILD": "5005",
                "PLACE": "A5",
                "LS": "x",
                "REC": "x",
            },  # 4: Fail! (not in PA)
            {
                "CHILD": "606",
                "PLACE": "A6",
                "LS": "x",
                "REC": "x",
            },  # 5: Fail! (missing vals in PA)
            {"CHILD": "77", "PLACE": "x", "LS": "x", "REC": "x"},  # 6 --ok
            {
                "CHILD": "8",
                "PLACE": "A6",
                "LS": "x",
                "REC": "X1",
            },  # 5: --ok (REC is X1)
        ]
    )

    test_placed = pd.DataFrame(
        [
            {
                "CHILD": "11",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
            },  # 0: Fail!
            {
                "CHILD": "202",
                "DATE_PLACED_CEASED": "01/02/2003",
                "REASON_PLACED_CEASED": "invalid dont matter",
            },  # 1
            {
                "CHILD": "3003",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
            },  # 2: E12 in EP --ok
            {
                "CHILD": "40004",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
            },  # 3: E12 in EP  --ok
            {
                "CHILD": "606",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
            },  # 4: Fail! (missing vals in PA)
        ]
    )

    fake_dfs = {
        "Episodes": test_episodes,
        "PlacedAdoption": test_placed,
    }

    result = validate(fake_dfs)

    assert result == {
        "PlacedAdoption": [0, 4],
        "Episodes": [0, 4, 5],
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW13STG1",
    message="The reason for social worker change is either not valid or has not been entered.",
    affected_fields=["SW_REASON"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]
        df["index"] = df.index

        valid_reasons = [
            "MANAGE",
            "FCONTA",
            "LEFTRL",
            "ORGRST",
            "TSPROC",
            "ABSENC",
            "CHCHAN",
            "PCCHAN",
            "SWDIED",
            "OTHERS",
        ]

        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")

        df.sort_values(["CHILD", "SW_DECOM", "SW_DEC"], ascending=True, inplace=True)

        index_of_first = df.drop_duplicates("CHILD", keep="first")

        allowed_nans = index_of_first[index_of_first["SW_REASON"].isna()][
            "index"
        ].tolist()

        error_rows = df[
            (~df["SW_REASON"].isin(valid_reasons) | df["SW_REASON"].isna())
        ]["index"].tolist()
        error_rows = set(error_rows) - set(allowed_nans)

        return {"SWEpisodes": list(error_rows)}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "SW_DECOM": "01/04/2000",
                "SW_DEC": "02/04/2000",
                "SW_REASON": "OTHERS",
            },  # 0 pass
            {
                "CHILD": "2",
                "SW_DECOM": "01/04/2000",
                "SW_DEC": "02/04/2000",
                "SW_REASON": "OTHERS",
            },  #  1 pass
            {
                "CHILD": "2",
                "SW_DECOM": "03/04/2000",
                "SW_DEC": "04/04/2000",
                "SW_REASON": pd.NA,
            },  # 2 fail
            {
                "CHILD": "3",
                "SW_DECOM": "03/04/1998",
                "SW_DEC": "04/04/2000",
                "SW_REASON": pd.NA,
            },  # 3 pass
            {
                "CHILD": "3",
                "SW_DECOM": "05/04/2000",
                "SW_DEC": "06/04/2000",
                "SW_REASON": pd.NA,
            },  # 4 fail
            {
                "CHILD": "4",
                "SW_DECOM": "05/04/1998",
                "SW_DEC": "31/03/2000",
                "SW_REASON": pd.NA,
            },  # 5 pass - first in report data
            {
                "CHILD": "4",
                "SW_DECOM": "05/04/2000",
                "SW_DEC": "31/04/2000",
                "SW_REASON": pd.NA,
            },  # 6 fail second in report data
        ]
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [2, 4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="365",
    message="Any individual short- term respite placement must not exceed 17 days.",
    affected_fields=["LS", "DECOM", "DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    episodes = dfs["Episodes"]
    collection_end_str = dfs["metadata"]["collection_end"]

    episodes.loc[episodes["DEC"].isna(), "DEC"] = collection_end_str
    episodes["DECOM"] = pd.to_datetime(
        episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    episodes["DEC"] = pd.to_datetime(
        episodes["DEC"], format="%d/%m/%Y", errors="coerce"
    )

    over_17_days = episodes["DEC"] > episodes["DECOM"] + pd.DateOffset(days=17)
    error_mask = (episodes["LS"] == "V3") & over_17_days

    return {"Episodes": episodes.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    test_eps = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DECOM": "01/01/2000",
                "DEC": "01/01/2001",
                "LS": "V3",
            },  # 0 Fail!
            {"CHILD": 101, "DECOM": "01/01/2001", "DEC": "20/12/2001", "LS": "X"},  # 1
            {
                "CHILD": 101,
                "DECOM": "20/12/2001",
                "DEC": "12/01/2002",
                "LS": "V3",
            },  # 2 Fail!
            {
                "CHILD": 2002,
                "DECOM": "01/01/2000",
                "DEC": "10/01/2000",
                "LS": "V3",
            },  # 3
            {"CHILD": 2002, "DECOM": "10/01/2000", "DEC": "23/01/2001", "LS": "X"},  # 4
            {
                "CHILD": 2002,
                "DECOM": "01/01/2002",
                "DEC": "18/01/2002",
                "LS": "V3",
            },  # 5 pass -17 days
            {
                "CHILD": 2002,
                "DECOM": "01/02/2002",
                "DEC": "19/02/2002",
                "LS": "V3",
            },  # 6 Fail! - 18 days
            {
                "CHILD": 2002,
                "DECOM": "11/01/2002",
                "DEC": "25/01/2002",
                "LS": "J2",
            },  # 7
            {"CHILD": 30003, "DECOM": "25/01/2002", "DEC": "10/03/2001", "LS": "J2"},
            # 8 pass - bigger problems: DEC < DECOM
            {
                "CHILD": 30003,
                "DECOM": "25/01/2002",
                "DEC": pd.NA,
                "LS": "V3",
            },  # 9 Fail! (DEC.isna --> collection_end )
            {
                "CHILD": 30003,
                "DECOM": pd.NA,
                "DEC": "25/01/2002",
                "LS": "V3",
            },  # 10 pass - bigger problems: no DECOM
            {
                "CHILD": 30003,
                "DECOM": pd.NA,
                "DEC": pd.NA,
                "LS": "V3",
            },  # 11 pass - bigger problems: no dates
        ]
    )

    test_meta = {"collection_end": "01/04/2002"}

    test_dfs = {"Episodes": test_eps, "metadata": test_meta}

    result = validate(test_dfs)

    assert result == {"Episodes": [0, 2, 6, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="147",
    message="Date episode ceased is not a valid date.",
    affected_fields=["DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        mask = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        ).notna()

        na_location = episodes["DEC"].isna()

        validation_error_mask = ~mask & ~na_location
        validation_error_locations = episodes.index[validation_error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DEC": ["01/01/2021", "19/02/2010", "38/04/2019", "01/01/19", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="521",
    message="Date of local authority's decision (LA) that adoption is in the best interests of the child (date should be placed) must be on or prior to the date the child is placed for adoption.",
    affected_fields=["PLACE", "DECOM", "DATE_INT"],
    tables=["Episodes", "AD1"],
)
def validate(dfs):
    if "Episodes" not in dfs or "AD1" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        ad1 = dfs["AD1"]
        code_list = ["A3", "A4", "A5", "A6"]
        # if PLACE is equal to A3, A4, A5 or A6 then placed-for-adoption = Y

        # to datetime
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        ad1["DATE_INT"] = pd.to_datetime(
            ad1["DATE_INT"], format="%d/%m/%Y", errors="coerce"
        )

        # prepare to merge
        episodes.reset_index(inplace=True)
        ad1.reset_index(inplace=True)
        merged = episodes.merge(ad1, how="left", on="CHILD", suffixes=["_eps", "_ad1"])

        # <DATE_INT> must be <= <DECOM> where <PLACED_FOR_ADOPTION> = 'Y'
        mask = merged["PLACE"].isin(code_list) & (merged["DATE_INT"] > merged["DECOM"])
        # error locations
        ad1_error_locs = merged.loc[mask, "index_ad1"]
        eps_error_locs = merged.loc[mask, "index_eps"]
        return {"Episodes": eps_error_locs.tolist(), "AD1": ad1_error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_ad1 = pd.DataFrame(
        {
            "DATE_INT": [
                "08/03/2020",
                "22/07/2020",
                "13/10/2021",
                "22/06/2020",
                pd.NA,
            ],
            "CHILD": [
                "111",
                "123",
                "333",
                "444",
                "678",
            ],
        }
    )
    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/01/2020",
                "PLACE": "A6",
            },  # 0 fail
            {
                "CHILD": "111",
                "DECOM": "01/11/2020",
                "PLACE": "A5",
            },  # 1 pass
            {
                "CHILD": "111",
                "DECOM": "22/01/2020",
                "PLACE": "X1",
            },  # 2 ignore PLACE not in list
            {
                "CHILD": "123",
                "DECOM": pd.NA,
                "PLACE": "A5",
            },  # 3 ignore DECOM is nan
            {
                "CHILD": "123",
                "DECOM": "11/01/2020",
                "PLACE": "X1",
            },  # 4 ignore PLACE not in list
            {
                "CHILD": "333",
                "DECOM": "01/01/2020",
                "PLACE": "A6",
            },  # 5 fail
            {
                "CHILD": "333",
                "DECOM": "22/01/2020",
                "PLACE": "X1",
            },  # 6 ignore PLACE not in list
            {
                "CHILD": "333",
                "DECOM": "11/01/2020",
                "PLACE": "U1",
            },  # 7 ignore PLACE not in list
            {
                "CHILD": "444",
                "DECOM": "22/06/2020",
                "PLACE": "A3",
            },  # 8 pass
            {
                "CHILD": "444",
                "DECOM": "11/01/2020",
                "PLACE": "X1",
            },  # 9 ignore PLACE not in list
            {
                "CHILD": "444",
                "DECOM": "01/01/2020",
                "PLACE": "A3",
            },  # 10 fail
            {
                "CHILD": "678",
                "DECOM": "01/01/2020",
                "PLACE": "A4",
            },  # 11 ignore DATE_INT is nan
        ]
    )
    fake_dfs = {"Episodes": fake_data_episodes, "AD1": fake_ad1}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 5, 10], "AD1": [0, 2, 3]}
